# -*- coding: utf-8 -*-

require 'date'
require 'csv'

require 'sidfmdlm_api'
require 'sidfm_controller_support'

class SidfmLicensesController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  # before_action :api_init, :except => [ :license_setting, :license_update ]

  private

  include SidfmControllerSupport
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license, :except => [ :license, :license_setting, :license_update ]
  before_action :check_roles

  MEMO_MAX_LEN = 100

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'register_host', 'toggle_host_delete', 'rename_host', 'update_memo', 'update_memo_bulk', 'update_refdate', 'update_refdate_bulk', 'update_categories', 'update_categories_bulk', 'update_applications', 'update_applications_bulk', 'load_offline_data' ].include?(action_name) and !@sidfm_roles.include?(:use) and !@role_admin then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    if [ 'license', 'license_setting', 'license_update', 'license_apikey_update', 'host_apikey_update', 'register_apikey_update', 'register_apikey_remove', 'import', 'import_apply', 'export' ].include?(action_name) and !@role_admin then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @user_category_ids = SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
    @user_host_ids = SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
    @cache_prefix = @role_admin ? '/license/admin/0' : "/license/user/#{@user_category_ids.join(",")}"
  end

  # get_hosts_by_lmのitemをCSVの1行分のitemに変換する。
  def item2csv(host)
    item = []
    item << (host[:sidfmd_host] ? host[:sidfmd_host].id : '').to_s.encode('UTF-16LE')
    item << host[:apikey].encode('UTF-16LE')
    item << host[:hostname].encode('UTF-16LE')
    item << (host[:sidfmd_host] ? host[:sidfmd_host].sidfmd_categories : []).collect {|x| x.name.encode('UTF-16LE') }.join(','.encode('UTF-16LE')).encode('UTF-16LE')
    item << (host[:sidfmd_host] ? (host[:sidfmd_host].refdate ? host[:sidfmd_host].refdate.strftime('%Y/%m/%d') : '') : '').encode('UTF-16LE')
    item << host[:flg_delete].to_s.encode('UTF-16LE')
    item << (host[:sidfmd_host] ? host[:sidfmd_host].is_agent.to_s : '').encode('UTF-16LE')
    item << (host[:sidfmd_host] ? (host[:sidfmd_host].memo ? host[:sidfmd_host].memo : '') : '').encode('UTF-16LE')
    if host[:sidfmd_host] then
      host[:sidfmd_host].sidfmd_host_apps.each do |app|
        if app.version_id then
          version = SidfmdVersion.where(:vid => app.version_id).first
          if version then
            item << version.fullname.encode('UTF-16LE')
          elsif app.product_id then
            product = SidfmdProduct.where(:pid => app.product_id).first
            if product then
              item << product.name.encode('UTF-16LE')
            end
          end
        elsif app.product_id then
          product = SidfmdProduct.where(:pid => app.product_id).first
          if product then
            item << product.name.encode('UTF-16LE')
          end
        end
      end
    end
    item
  end

  def check_filters
    @hosts[:filters][:category_id] = '0' unless @hosts[:filters][:category_id]
    @hosts[:filters][:hostname] = '' unless @hosts[:filters][:hostname]
  end

  REFDATE_INVALID = Date.new(4000,12,31).freeze
  REFDATE_NONE = Date.new(3000,12,31).freeze

  def apply_filters
    ret = @hosts[:hosts].find_all {|item| 
      # カテゴリ検索
      if @hosts[:filters][:category_id] != '0' then
        category_id = @hosts[:filters][:category_id].to_i
        if item[:sidfmd_host] then
          next false if category_id == -2
          categories = item[:sidfmd_host].sidfmd_categories
          categories = categories.where(:id => category_id) if category_id > 0
          categories_count = categories.count
          next false if (category_id == -1 and categories_count != 0)
          next false if (category_id > 0 and categories_count == 0)
        else
          next false if category_id != -2
        end
      end
      # ホスト名照合
      if @hosts[:filters][:hostname] != '' then
        next false unless item[:hostname].downcase[@hosts[:filters][:hostname].downcase]
      end
      true
    }
    case @hosts[:order]
    when 'hostname_asc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, x[:hostname].downcase ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'hostname_desc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, x[:hostname].downcase ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'category_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          cats = x[:sidfmd_host].sidfmd_categories.system
          if cats.size > 0 then
            [ x[:flg_delete] ? 1000 : 100, 100, cats.first.id ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'category_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          cats = x[:sidfmd_host].sidfmd_categories.system
          if cats.size > 0 then
            [ x[:flg_delete] ? 1000 : 100, 100, cats.first.id ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'os_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          os = x[:sidfmd_host].get_os
          if os and os.has_key?(:osname) then
            [ x[:flg_delete] ? 1000 : 100, 100, os[:osname] ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'os_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          os = x[:sidfmd_host].get_os
          if os and os.has_key?(:osname) then
            [ x[:flg_delete] ? 1000 : 100, 100, os[:osname] ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agentprds_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          queue = x[:sidfmd_host].sidfmd_queues.order(:created_at => :desc).first
          if queue then
            [ x[:flg_delete] ? 1000 : 100, queue.sidfmd_queue_pkgs.size + queue.sidfmd_queue_apps.size ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'agentprds_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          queue = x[:sidfmd_host].sidfmd_queues.order(:created_at => :desc).first
          if queue then
            [ x[:flg_delete] ? 1000 : 100, queue.sidfmd_queue_pkgs.size + queue.sidfmd_queue_apps.size ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'prds_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          lapps = x[:sidfmd_host].sidfmd_host_apps
          if lapps then
            [ x[:flg_delete] ? 1000 : 100, lapps.count ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'prds_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          lapps = x[:sidfmd_host].sidfmd_host_apps
          if lapps then
            [ x[:flg_delete] ? 1000 : 100, lapps.count ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'refdate_asc' then 
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].refdate then
            [ x[:flg_delete] ? 1000 : 100, x[:sidfmd_host].refdate ]
          else
            [ x[:flg_delete] ? 1000 : 100, REFDATE_NONE ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, REFDATE_INVALID ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'refdate_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].refdate then
            [ x[:flg_delete] ? 1000 : 100, x[:sidfmd_host].refdate ]
          else
            [ x[:flg_delete] ? 1000 : 100, REFDATE_NONE ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, REFDATE_INVALID ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'memo_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].memo then
            [ x[:flg_delete] ? 1000 : 100, 100 ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'memo_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].memo then
            [ x[:flg_delete] ? 1000 : 100, 100 ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'manual_asc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and !x[:sidfmd_host].is_agent ? 100 : 1000), x[:hostname] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'manual_desc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and !x[:sidfmd_host].is_agent ? 100 : 1000), x[:hostname] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agent_asc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and x[:sidfmd_host].is_agent ? 100 : 1000), x[:hostname] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agent_desc' then
      temp_ev = proc {|x|
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and x[:sidfmd_host].is_agent ? 100 : 1000), x[:hostname] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    else
      # どれにも該当しない場合はホスト名検索とする。
      ret.sort! {|x, y| x[:hostname].downcase <=> y[:hostname].downcase }
    end
    ret
  end

  def create_content
    ret = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
    if ret then
      if @role_admin then
        ret[:hosts].delete_if {|x| x[:sidfmd_host].nil? }
      else
        ret[:hosts].delete_if {|x| x[:sidfmd_host].nil? or !@user_host_ids.include?(x[:sidfmd_host].id) }
      end
    end
    ret
  end

  public

  def index
    pages_params = params.fetch(:pages, {}).permit(:offset, :per_page, :filters, :order, :checked, :all_checked)
    @hosts = {}
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 100).to_i
    @hosts[:order] = (pages_params[:order] || 'hostname_asc')
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = ((pages_params[:all_checked] || '') == 'true')
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
  end

  def index_r
    pages_params = params.fetch(:pages, {}).permit(:offset, :per_page, :filters, :order, :checked, :all_checked)

    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
    @register_apikey = Sidfm.sidfmdlm_register_apikey
    @hosts = Rails.cache.fetch(@cache_prefix + '/hosts/0', SidfmControllerSupport::CACHE_OPTIONS) do
      create_content
    end
    if @hosts.nil? then
      Rails.cache.clear
      redirect_to :action => :license, :notice => 'ライセンスが不正です。' and return
    end
    if !@sidfm_roles.include?(:admin) then
      @hosts[:hosts].delete_if {|x| x[:flg_delete] }
    end
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 100).to_i
    @hosts[:order] = (pages_params[:order] || 'hostname_asc')
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = ((pages_params[:all_checked] || '') == 'true')
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    check_filters
    @hosts[:hosts_filtered] = apply_filters
    @all_hostnames =  @hosts[:hosts_filtered].map{|data| data[:hostname]}
    @hosts[:content] = @hosts[:hosts_filtered].slice(@hosts[:offset], @hosts[:per_page])
    @hosts[:content] = [] unless @hosts[:content]
    @hosts[:total_size] = @hosts[:hosts_filtered].size
 
    @system_categories = SidfmdCategory.system
    @not_system_categories = @role_admin ? SidfmdCategory.not_system : SidfmdCategory.where(:id => @user_category_ids)
  end

  def license
    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
    @register_apikey = Sidfm.sidfmdlm_register_apikey
    #filtered_params = params.fetch(:params, {}).permit(:offset, :limit, :hostname, :'sort[hostname]', :'sort[flg_delete]', :'sort[created_at]')
    @hosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
  end

  def license_setting
    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
  end

  def license_update
    filtered_params = params.fetch(:params, {}).permit(:license_key, :license_password)
    @license_key = filtered_params[:license_key]
    license_password = filtered_params[:license_password]
    begin
      if !SidfmdHost.auth_api(@license_key, license_password) then
        redirect_to :action => :license_setting, :notice => 'ライセンス認証に失敗しました。' and return
      end
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :license_setting, :notice => 'ライセンスサーバへのアクセスに失敗しました。' and return
    end
    Rails.cache.clear

    redirect_to :action => :license
  end

  def license_apikey_update
    res = SidfmdHost.license_apikey_update
    if res.respond_to?(:[]) and res[:success] then
      redirect_to :action => :license and return
    end
    redirect_to :action => :license, :notice => 'APIキーの更新に失敗しました。'
  end

  # ホストAPIキーの更新
  # deprecated:使用せず。
  def host_apikey_update
    filtered_params = params.fetch(:params, {}).permit(:id, :hostname)
    host_id = filtered_params[:id]
    hostname = filtered_params[:hostname]
    if host_id and hostname then
      SidfmdHost.transaction do
        res = SidfmdHost.update_host_apikey(host_id.to_i)
        if res.respond_to?(:[]) and res[:success] then
          host = SidfmdHost.where(:hostname => hostname).first
          if host then
            host.update_attribute :host_apikey, res[:apikey]
          else
            res = SidfmdHost.establish_consistency
          end
          redirect_to :action => :index and return
        end
      end
    end
    redirect_to :action => :index_r, :notice => 'APIキーの更新に失敗しました。'
  end

  # 削除状態の変更
  def toggle_host_delete
    filtered_params = params.fetch(:params, {}).permit(:id, :hostname, :dump, :rule_info)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    host_id = filtered_params[:id]
    hostname = filtered_params[:hostname]
    if host_id and hostname then
      host = SidfmdHost.where(:hostname => hostname).first
      if host then
        if !@role_admin and !@user_host_ids.include?(host.id) then
          dump[:notice] = '自分の属するカスタムカテゴリのホスト以外は操作できません。'
          redirect_to dump and return
        end
      end
      res = nil
      begin
        rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
        res = SidfmdHost.toggle_host_delete(host_id.to_i)
        logger.info(res.inspect)
        unless res then
          dump[:notice] = 'ライセンスサーバとの通信に失敗しました。'
          redirect_to dump and return
        end
        unless res[:success] then
          dump[:notice] = 'ライセンス数を超過するため、操作に失敗しました。'
          redirect_to dump and return
        end
      rescue Exception => e
        log_exception(e)
        dump[:notice] = 'ライセンスサーバへのアクセスに失敗しました。'
        redirect_to dump and return
      end
      host = nil
      SidfmdHost.transaction do
        host = SidfmdHost.where(:hostname => hostname).first
        if host then
          host.lock!
          hostid = host.id
          new_flg_delete = !host[:flg_delete]
          host.update_attribute :flg_delete, new_flg_delete
          if new_flg_delete then
            # データを削除する。Redmineのチケットは削除しない（VM上からは直接参照できない）。
            host.last_checked_at = nil
            host.memo = nil
            host.is_agent = false
            host.save
            # カテゴリの削除
            SidfmdCategoriesHost.where(:sidfmd_host_id => hostid).delete_all
            # プロダクトの削除
            SidfmdHostApp.where(:sidfmd_host_id => hostid).delete_all
            # キューの削除
            sidfmd_queue_ids = SidfmdQueue.where(:sidfmd_host_id => hostid).collect {|x| x.id }.uniq
            sidfmd_queue_ids.each do |sidfmd_queue_id|
              SidfmdQueueDist.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueuePkg.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueApp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueLapp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueWin.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              sidfmd_queue_vul_ids = SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).collect {|x| x.id }.uniq
              sidfmd_queue_vul_ids.each do |sidfmd_queue_vul_id|
                SidfmdQueueVulCve.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulSri.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffprod.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffver.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffpkg.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              end
              SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            end
            SidfmdQueue.where(:id => sidfmd_queue_ids).delete_all
            # チケット（脆弱性）の削除
            sidfmd_issue_ids = SidfmdIssue.where(:sidfmd_host_id => hostid).collect {|x| x.id }.uniq
            sidfmd_issue_ids.each do |sidfmd_issue_id|
              SidfmdIssueCve.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueSri.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffprod.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffver.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffpkg.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            end
            SidfmdIssue.where(:id => sidfmd_issue_ids).delete_all
            #ルールまたはルールの条件削除
            delete_rule_ids = rule_info[:delete_rule_ids]
            delete_cond_ids = rule_info[:delete_cond_ids]
            delete_host_cond_ids = rule_info[:delete_host_cond_ids]
            if delete_rule_ids && delete_rule_ids.size > 0 then
              SidfmdRule.where(:id => delete_rule_ids).destroy_all
            end 
            if delete_cond_ids && delete_cond_ids.size > 0 then
              SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
            end
            if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
              SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
            end
          end
        end
      end
      unless host then
        SidfmdHost.transaction do
          res = SidfmdHost.establish_consistency
          logger.info("SidfmdHost.establish_consistensy=#{res.inspect}")
        end
      end
      Rails.cache.clear
      redirect_to dump and return
    end
    dump[:notice] = '削除状態の変更に失敗しました。'
    redirect_to dump and return
  end

  # ホスト名変更
  def rename_host
    filtered_params = params.fetch(:params, {}).permit(:hostid, :hostname, :old_hostname, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostid = filtered_params[:hostid]
    hostname = filtered_params[:hostname]
    old_hostname = filtered_params[:old_hostname]
    unless hostid or hostname or old_hostname then
      dump[:notice] = 'パラメータが不正です。'
      redirect_to dump and return
    end
    if hostid.size == 0 or hostname.size == 0 or old_hostname.size == 0 then
      dump[:notice] = 'パラメータが不正です。'
      redirect_to dump and return
    end
    if hostid and hostname then
      unless @role_admin then
        host = SidfmdHost.where(:hostname => old_hostname).first
        if host then
          if !@user_host_ids.include?(host.id) then
            dump[:notice] = '自分の属するカスタムカテゴリのホスト以外は操作できません。'
            redirect_to dump and return
          end
        else
          dump[:notice] = 'ホストが存在しません。'
          redirect_to dump and return
        end
      end
      SidfmdHost.transaction do
        res = SidfmdHost.rename_host(hostid, hostname)
        if res.respond_to?(:[]) then
          if res[:success] then
            host = SidfmdHost.where(:hostname => old_hostname).first
            if host then
              host.update_attribute :hostname, hostname
            else
              res = SidfmdHost.establish_consistency
            end
            Rails.cache.clear
            redirect_to dump and return
          else
            dump[:notice] = '登録済のホスト名、もしくは不正な文字を指定したため、操作に失敗しました。'
            redirect_to dump and return
          end
        else
          dump[:notice] = 'ライセンスサーバとの通信に失敗しました。'
          redirect_to dump and return
        end
      end
    end
    dump[:notice] = 'ホスト名の変更に失敗しました。'
    redirect_to dump and return
  end

  # メモ更新
  def update_memo
    filtered_params = params.fetch(:params, {}).permit(:hostname, :memo, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostname = filtered_params[:hostname]
    memo = filtered_params[:memo]
    unless hostname or memo then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostname.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if memo.length > MEMO_MAX_LEN then
      redirect_to :action => :index_r, :notice => 'メモに入力可能な最大文字数を超えています' and return
    end
    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then
        dump[:notice] = 'ホストが存在しません。'
        redirect_to dump and return
      end
      host.update_attribute :memo, memo.size > 0 ? memo : nil
      Rails.cache.clear
    end
    redirect_to dump
  end

  # メモ一括更新
  def update_memo_bulk
    filtered_params = params.fetch(:params, {}).permit(:hostnames, :memo, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostnames = filtered_params[:hostnames]
    memo = filtered_params[:memo]
    unless hostnames or memo then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostnames.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    memo = nil if memo.size == 0
    SidfmdHost.transaction do
      hostnames.each do |hostname|
        host = SidfmdHost.where(:hostname => hostname).first
        if host then
          host.update_attribute :memo, memo
        end
      end
      Rails.cache.clear
    end
    redirect_to dump
  end

  # 基準日更新
  def update_refdate
    filtered_params = params.fetch(:params, {}).permit(:hostname, :refdate, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostname = filtered_params[:hostname]
    refdate = filtered_params[:refdate]
    unless hostname or refdate then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostname.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then 
        dump[:notice] = 'ホストが存在しません。'
        redirect_to dump and return
      end
      if refdate.size > 0 then
        begin
          refdate = Date.parse(refdate) if refdate.size > 0
        rescue ArgumentError => e
          log_exception(e)
          dump[:notice] = '不正な日付です。'
          redirect_to dump and return
        end
      else
        refdate = nil
      end
      begin
        host.update_attribute :refdate, refdate
        Rails.cache.clear
      rescue ActiveRecord::StatementInvalid => e
        log_exception(e)
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    redirect_to dump
  end

  # 基準日一括更新
  def update_refdate_bulk
    filtered_params = params.fetch(:params, {}).permit(:hostnames, :refdate, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostnames = filtered_params[:hostnames]
    refdate = filtered_params[:refdate]
    unless hostnames or refdate then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostnames.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if refdate.size > 0 then
      begin
        refdate = Date.parse(refdate) if refdate.size > 0
      rescue ArgumentError => e
        log_exception(e)
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    else
      refdate = nil
    end
    begin
      SidfmdHost.transaction do
        hostnames.each do |hostname|
          host = SidfmdHost.where(:hostname => hostname).first
          if host then
            host.update_attribute :refdate, refdate
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    redirect_to dump
  end

  # カテゴリ更新
  def update_categories
    filtered_params = params.fetch(:categories, {}).permit(:sidfmdhostid, :hostname, :selected_category_ids, :system_category_id, :dump, :rule_info)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
      rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    sidfmdhostid = filtered_params[:sidfmdhostid].to_i
    hostname = filtered_params[:hostname]
    selected_category_ids = filtered_params[:selected_category_ids]
    system_category_id = filtered_params[:system_category_id].to_i
    if sidfmdhostid == 0 or system_category_id == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if !hostname or !selected_category_ids then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if !@role_admin then
      if selected_category_ids.size == 0 then
        redirect_to :action => :index_r, :notice => '少なくとも一つはカスタムカテゴリを指定する必要があります。' and return
      else
        diff = selected_category_ids & @user_category_ids
        if diff.size == 0 then
          redirect_to :action => :index_r, :notice => '自分の属するカスタムカテゴリ以外を指定することはできません。' and return
        end
        selected_category_ids = diff
      end
    end
    begin
      SidfmdHost.transaction do
        host = SidfmdHost.where(:id => sidfmdhostid).first
        unless host then
          host = SidfmdHost.where(:hostname => hostname).first
          unless host then
            dump[:notice] = '対象ホストが見つかりません。'
            redirect_to dump and return
          end
        end
        #関連するルールの削除・変更
        delete_rule_ids = rule_info[:delete_rule_ids]
        delete_cond_ids = rule_info[:delete_cond_ids]
        delete_host_cond_ids = rule_info[:delete_host_cond_ids]
        if delete_rule_ids && delete_rule_ids.size > 0 then
          SidfmdRule.where(:id => delete_rule_ids).destroy_all
        end
        if delete_cond_ids && delete_cond_ids.size > 0 then
          SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
        end
        if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
          SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
        end

        res = host.modify_categories(system_category_id, selected_category_ids,
                                     @role_admin ? nil : @user_category_ids)
        unless res then
          raise StandardError, 'modify_categories failed'
        end
      end
      Rails.cache.clear
    rescue Exception => e
      log_exception(e)
      dump[:notice] = 'カテゴリ更新に失敗しました。'
      redirect_to dump and return
    end
    Rails.cache.clear
    redirect_to dump
  end

  # カテゴリ一括変更
  def update_categories_bulk
    filtered_params = params.fetch(:categories, {}).permit(:hostnames, :selected_category_ids, :system_category_id, :dump, :rule_info)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
      rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostnames = filtered_params[:hostnames]
    unless hostnames then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    system_category_id = filtered_params[:system_category_id]
    unless system_category_id then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    selected_category_ids = filtered_params[:selected_category_ids]
    unless selected_category_ids then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      SidfmdHost.transaction do
        #関連するルールの削除・変更
        delete_rule_ids = rule_info[:delete_rule_ids]
        delete_cond_ids = rule_info[:delete_cond_ids]
        delete_host_cond_ids = rule_info[:delete_host_cond_ids]
        if delete_rule_ids && delete_rule_ids.size > 0 then
          SidfmdRule.where(:id => delete_rule_ids).destroy_all
        end
        if delete_cond_ids && delete_cond_ids.size > 0 then
          SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
        end
        if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
          SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
        end

        SidfmdHost.where(:hostname => hostnames).each do |host|
          res = host.modify_categories(system_category_id, selected_category_ids)
          unless res then
            logger.info('modify_categories failed')
            raise StandardError
          end
        end
      end
      Rails.cache.clear
    rescue Exception => e
      log_exception(e)
      dump[:notice] = 'カテゴリ更新に失敗しました。'
      redirect_to dump and return
    end
    redirect_to dump
  end

  def register_apikey_update
    res = Sidfm.sidfmdlm_register_apikey_update
    if res then
      redirect_to :action => :license
      return
    end
    redirect_to :action => :license, :notice => 'APIキーの登録・更新に失敗しました。'
  end

  def register_apikey_remove
    Sidfm.sidfmdlm_register_apikey = nil
    unless Sidfm.sidfmdlm_register_apikey then
      redirect_to :action => :license
      return
    end
    redirect_to :action => :license, :notice => 'APIキーの削除に失敗しました。'
  end

  # ホストを登録する。
  def register_host
    f_params = params.fetch(:params, {}).permit(:hostname, :category_id, :selected, :selected_category_ids, :refdate, :memo)
    hostname = f_params[:hostname]
    if /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
      redirect_to :action => :index_r, :notice => 'ホスト名が不正です。' and return
    end
    category_id = f_params[:category_id].to_i
    unless category_id > 0 then
      redirect_to :action => :index_r, :notice => 'システムカテゴリが指定されていません。' and return
    end
    selected_category_ids = f_params[:selected_category_ids]
    unless selected_category_ids then
      redirect_to :action => :index_r, :notice => 'カスタムカテゴリが設定されていません。' and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => 'カスタムカテゴリが不正です。' and return
    end
    unless selected_category_ids then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    if !@role_admin then
      if selected_category_ids.size == 0 then
        redirect_to :action => :index_r, :notice => '少なくとも一つはカスタムカテゴリを指定する必要があります。' and return
      else
        diff = selected_category_ids & @user_category_ids
        if diff.size == 0 then
          redirect_to :action => :index_r, :notice => '自分の属するカスタムカテゴリ以外を指定することはできません。' and return
        end
        selected_category_ids = diff
      end
    end
    selected = f_params[:selected]
    unless selected then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    begin
      selected = JSON.parse(selected)
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    refdate = f_params[:refdate]
    unless refdate then
      redirect_to :action => :index_r, :notice => '基準日が不正です。' and return
    end
    if refdate.size > 0 then
      begin
        refdate = Date.parse(refdate) if refdate.size > 0
      rescue ArgumentError => e
        log_exception(e)
        redirect_to :action => :index_r, :notice => '不正な日付です。' and return
      end
    else
      redirect_to :action => :index_r, :notice => '基準日は必須です。' and return
    end
    memo = f_params[:memo]
    unless memo then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    if memo.length > MEMO_MAX_LEN then
      redirect_to :action => :index_r, :notice => 'メモに入力可能な最大文字数を超えています' and return
    end    
    memo = nil if memo.size == 0

    # 入力チェック
    if hostname.size == 0 then
      redirect_to :action => :index_r, :notice => 'ホスト名が入力されていません。'
      return
    end
    if category_id <= 0 then
      redirect_to :action => :index_r, :notice => 'カテゴリが不正です。'
      return
    end

    # 既に登録されている場合はreject
    # 削除済ならそのまま通り（有効になる）。
    old_host = SidfmdHost.where(:hostname => hostname).where(:flg_delete => false)
    if old_host and old_host.count > 0 then
      redirect_to :action => :index_r, :notice => '同名のホストが登録されています。' and return
    end

    @hosts = SidfmdHost.get_hosts_by_lm(:limit => 0)

    # ライセンスチェック
    # なお、ライセンスサーバ側でもチェックが行われるため、このチェックを除外しても
    # ライセンス数を超えてホストの登録を行うことはできない。
    if @hosts and @hosts[:license_number] <= @hosts[:used_license_number] then
      redirect_to :action => :index_r, :notice => 'ホスト数がライセンスの上限に達しています。'
      return
    end

    # ライセンスサーバとVMでホストの登録を行う。
    temp = SidfmdHost.register_host(hostname, nil, false)
    if temp.nil? then
      redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。' and return
    end

    # 作成されたホストを取得
    host = SidfmdHost.where(:hostname => hostname).first
    if host.nil? then
      redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。' and return
    end

    begin
      SidfmdHost.transaction do
        # カテゴリ登録
        SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                    :sidfmd_category_id => category_id)
        res = host.modify_categories(category_id, selected_category_ids)
        unless res then
          raise StandardError
        end

        # 基準日登録
        host.update_attribute :refdate, refdate

        # メモ登録
        host.update_attribute :memo, memo

        # アプリ登録
        host.sidfmd_host_apps = []
        host.save
        selected.each_pair do |k, v|
          v.each do |vv|
            pid = vv['pid'].to_i
            vid = vv['vid'].to_i
            SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                 :product_id => pid,
                                 :version_id => vid)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      redirect_to :action => :index, :notice => '不正な日付です。' and return
    rescue StandardError => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'カテゴリの更新に失敗しました。' and return
    end
    redirect_to :action => :index
  end

  # 設定されているアプリケーションを更新する。
  def update_applications
    f_params = params.fetch(:params, {}).permit(:hostname, :selected, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostname = f_params[:hostname]
    selected = JSON.parse(f_params[:selected])

    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then
        dump[:notice] = 'ホストが存在しません'
        redirect_to dump and return
      end
      host.sidfmd_host_apps = []
      host.save

      selected.each_pair do |k, v|
        v.each do |vv|

          pid = vv['pid'].to_i
          vid = vv['vid'].to_i

          SidfmdHostApp.create(:sidfmd_host_id => host.id,
                               :product_id => pid,
                               :version_id => vid)
        end
      end
    end
    Rails.cache.clear

    redirect_to dump
  end

  # プロダクト一括更新
  def update_applications_bulk
    filtered_params = params.fetch(:params, {}).permit(:hostnames, :selected, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostnames = filtered_params[:hostnames]
    selected = filtered_params[:selected]
    unless hostnames or selected then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e 
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      selected = JSON.parse(selected)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    SidfmdHost.transaction do
      hostnames.each do |hostname|
        host = SidfmdHost.where(:hostname => hostname).first
        if host then
          host.sidfmd_host_apps = []
          host.save
          selected.each_pair do |k, v|
            v.each do |vv|
              pid = vv['pid'].to_i
              vid = vv['vid'].to_i
              SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                   :product_id => pid,
                                   :version_id => vid)
            end
          end
        end
      end
    end
    Rails.cache.clear
    redirect_to dump
  end

  # ホスト一覧のインポートを行い、操作予定の内容を出力する
  def import
    filtered_params = params.fetch(:params, {}).permit(:import_file)
    import_file = filtered_params[:import_file]
    unless import_file then
      redirect_to :action => :index_r, :notice => 'ファイルが選択されておりません。'
      return
    end
    lmhosts = nil
    begin
      lmhosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => 'ライセンスサーバへのアクセスに失敗しました。'
      return
    end
    unless lmhosts and lmhosts[:success] then
      redirect_to :action => :index_r, :notice => 'ホスト一覧の取得に失敗しました。'
      return
    end
    @hosts = lmhosts[:hosts]
    import_file = import_file.read
    unless import_file then
      redirect_to :action => :index_r, :notice => '送信されたCSVファイルが不正です。'
      return
    end
    if import_file.getbyte(0) == 255 and import_file.getbyte(1) == 254 then
      import_file = import_file[2, import_file.length - 2]
    end
    begin
      import_file = import_file.force_encoding('UTF-16LE').encode('UTF-8')
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => "送信されたCSVファイルが不正(#{e.message})です。" and return
    end
    csv = CSV.parse(import_file, :col_sep => "\t", :headers => :first_row)
    @actions = []
    @import_data = []
    new_tokens = []
    new_ht = []
    csv.each do |line|
      item = {}
      item[:host_id] = line['ホストID']
      item[:token] = line['トークン']
      new_tokens << item[:token] if item[:token]
      item[:hostname] = line['ホスト名']
      new_ht << [ item[:hostname], item[:token] ] if (item[:hostname] and item[:token])
      item[:categories] = line['カテゴリ']
      item[:refdate] = line['基準日']
      item[:flg_delete] = line['削除']
      item[:is_agent] = line['Agentフラグ']
      item[:memo] = line['メモ']
      item[:products] = []
      for i in 8..(line.size - 1) do
        item[:products] << line[i]
      end
      item[:products].uniq!
      item[:products].compact!
      @import_data << item
      # 同じhostnameのホストが存在していたらそのhost情報、存在しなければnil
      same_hostname = @hosts.find {|x| x[:hostname] == item[:hostname] }
      # 新規登録・復元のチェック
      if item[:token].nil? or item[:token] == '' then
        if same_hostname.nil? then
          if item[:hostname] and item[:hostname].size > 0 then
            # 新規登録
            @actions << {
              :action => 'create_new_host',
              :action_name => '新規作成',
              :new_data => item,
              :old_data => nil,
              :options => nil,
            }
            next
          end
        end
      end
      if same_hostname and same_hostname[:flg_delete] and
          item[:flg_delete] and item[:flg_delete].downcase == 'false' then
        # 復元
        @actions << {
          :action => 'undelete_host',
          :action_name => '復元',
          :new_data => item,
          :old_data => same_hostname,
          :options => nil,
        }
        next
      end
      same_token = @hosts.find {|x| x[:apikey] == item[:token]}
      # 更新のチェック
      if same_token and same_token[:sidfmd_host] then
        target = []
        # ホスト名比較
        if item[:hostname] != same_token[:hostname] then
          target << {
            :action => 'rename_host',
            :action_name => 'ホスト名変更',
            :old => same_token[:hostname],
            :new => item[:hostname],
          }
        end
        # カテゴリ比較
        i_categories = item[:categories] ? item[:categories].split(/,/) : []
        s_categories = same_token[:sidfmd_host].sidfmd_categories.collect {|x| x.name }
        if (i_categories - s_categories).size != 0 or
            (s_categories - i_categories).size != 0 then
          target << {
            :action => 'category_change',
            :action_name => 'カテゴリ変更',
            :old => s_categories,
            :new => i_categories,
          }
        end
        # 基準日比較
        i_refdate = item[:refdate].nil? ? '' : item[:refdate]
        s_refdate = same_token[:sidfmd_host].refdate
        if s_refdate then
          s_refdate = s_refdate.strftime('%Y/%m/%d')
        else
          s_refdate = ''
        end
        ev_i_refdate = ''
        if i_refdate.size > 0 then
          begin
            ev_i_refdate = Date.parse(i_refdate)
            if ev_i_refdate then
              ev_i_refdate = ev_i_refdate.strftime('%Y/%m/%d')
            end
          rescue Exception => e
            log_exception(e)
          end
        end
        if ev_i_refdate != s_refdate then
          target << {
            :action =>'refdate_change',
            :action_name => '基準日変更',
            :old => s_refdate,
            :new => i_refdate,
          }
        end
        # メモ比較
        i_memo = item[:memo]
        s_memo = same_token[:sidfmd_host].memo
        if (item[:memo] ? item[:memo] : '') != (same_token[:sidfmd_host].memo ? same_token[:sidfmd_host].memo : '') then
          if i_memo and i_memo.size > MEMO_MAX_LEN then
            redirect_to :action => :index_r, :notice => "メモのサイズが100文字を超えています。"
            return
          end
          target << {
            :action => 'memo_change',
            :action_name => 'メモ変更',
            :old => s_memo,
            :new => i_memo,
          }
        end
        # 製品比較
        same_token_products = []
        same_token[:sidfmd_host].sidfmd_host_apps.each do |app|
          if app.version_id then
            version = SidfmdVersion.where(:vid => app.version_id).first
            if version then
              same_token_products << version.fullname
            elsif app.product_id then
              product = SidfmdProduct.where(:pid => app.product_id).first
              if product then
                same_token_products << product.name
              end
            end
          elsif app.product_id then
            product = SidfmdProduct.where(:pid => app.product_id).first
            if product then
              same_token_products << product.name
            end
          end
        end
        same_token_products.compact!
        same_token_products.uniq!
        if (item[:products] - same_token_products).size != 0 or
            (same_token_products - item[:products]).size != 0 then
          target << {
            :action => 'product_change',
            :action_name => '製品変更',
            :old => same_token_products,
            :new => item[:products],
          }
        end
        if target.size > 0 then
          @actions << {
            :action => 'modify_host',
            :action_name => '設定変更',
            :old_data => item,
            :new_data => same_token,
            :options => target,
          }
        end
      end
    end
    # 削除チェック
    old_tokens = []
    old_ht = []
    @hosts.each do |host|
      if host[:sidfmd_host] then
        apikey = host[:sidfmd_host].host_apikey
        old_tokens << apikey if apikey
        old_ht << [ host[:hostname], apikey ] if (host[:hostname] and apikey)
      end
    end
    diff_tokens = old_tokens - new_tokens
    if diff_tokens.size > 0 then
      @actions << {
        :action => 'remove_host',
        :action_name => '削除',
        :old_data => diff_tokens.collect {|x| { :token => x, :host => @hosts.find {|y| x == y[:apikey] } } },
        :new_data => nil,
        :options => nil,
      }
    end
    #
    if @actions.size == 0 then
      redirect_to :action => :index_r, :notice => '変更が無いため、インポートを終了します。'
      return
    end
    # ライセンスチェック
    create_undelete_count = @actions.count {|x| [ 'create_new_host', 'undelete_host' ].include?(x[:action]) }
    delete_count = @actions.count {|x| 'remove_host' == x[:action] }
    create_count = create_undelete_count - delete_count
    if create_count > 0 then
      # ライセンス追加がある場合は、ライセンス数の確認を行う。
      lmhosts = SidfmdHost.get_hosts_by_lm(:limit => 0)
      if lmhosts and lmhosts[:license_number] - lmhosts[:used_license_number] < create_count then
        redirect_to :action => :index_r, :notice => 'ライセンス数を超えるため、インポートを実行することができません。'
        return
      end
    end
  end

  # インポート結果を実際に反映する。
  def import_apply
    f_params = params.fetch(:params, {}).permit(:hosts, :import_data, :actions)
    if !f_params[:hosts] or !f_params[:import_data] or !f_params[:actions] then
      redirect_to :action => :index_r, :notice => 'パラメータが不正です。'
      return
    end
    hosts = nil
    import_data = nil
    actions = nil
    begin
      hosts = JSON.parse(f_params[:hosts], :symbolize_names => true)
      import_data = JSON.parse(f_params[:import_data], :symbolize_names => true)
      actions = JSON.parse(f_params[:actions], :symbolize_names => true)
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => 'パラメータが不正です。' and return
    end
    actions.each do |action|
      case action[:action]
      when 'create_new_host' then # 新規登録
        # ライセンスチェック
        lmhosts = SidfmdHost.get_hosts_by_lm(:limit => 0)
        if lmhosts and lmhosts[:license_number] <= lmhosts[:used_license_number] then
          logger.info(lmhosts.inspect)
          redirect_to :action => :index_r, :notice => 'ホスト数がライセンスの上限に達しています。'
          return
        end

        # ライセンスサーバとVMでホストの登録を行う。
        temp = SidfmdHost.register_host(action[:new_data][:hostname], nil, false)
        if temp.nil? then
          redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。'
          return
        end

        # 作成されたホストを取得
        host = SidfmdHost.where(:hostname => action[:new_data][:hostname]).first
        unless host then
          redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。'
        end

        # カテゴリ登録
        if action[:categories] then
          SidfmdCategoriesHost.transaction do
            action[:categories].split(/,/).each do |category_name|
              cat = SidfmdCategory.where(:name => category_name).first
              if cat then
                SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                            :sidfmd_category_id => cat.id)
              end
            end
          end
        end

        # アプリ登録
        if action[:new_data][:products] then
          SidfmdHostApp.transaction do
            host.sidfmd_host_apps = []
            host.save
            action[:new_data][:products].each do |product_name|
              ver = SidfmdVersion.where(:fullname => product_name).first
              if ver then
                SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                     :product_id => ver.sidfmd_product.pid,
                                     :version_id => ver.vid)
              else
                prd = SidfmdProduct.where(:name => product_name).first
                if prd then
                  SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                       :product_id => prd.pid,
                                       :version_id => nil)
                end
              end
            end
          end
        end

        # メモ登録
        if action[:new_data][:memo] and action[:new_data][:memo].size > MEMO_MAX_LEN then
          redirect_to :action => :index_r, :notice => 'メモの長さが100文字を超えています。' and return
        end
        SidfmdHost.transaction do
          host.update_attribute :memo, action[:new_data][:memo]
          host.save
        end
      when 'undelete_host' then # 復元（削除状態のものを復元する）
        host_id = action[:old_data][:id].to_i
        hostname = action[:old_data][:hostname]
        # ライセンスサーバの更新
        host = nil
        if host_id > 0 and hostname and hostname.size > 0 then
          SidfmdHost.transaction do
            res = SidfmdHost.toggle_host_delete(host_id)
            if res then
              if res[:success] then
                host = SidfmdHost.where(:hostname => hostname).first
                if host then
                  host.update_attribute :flg_delete, !host[:flg_delete]
                else
                  res = SidfmdHost.establish_consistency
                end
              else
                redirect_to :action => :index_r, :notice => 'ライセンス数を超過するため、操作に失敗しました。' and return
              end
            else
              redirect_to :action => :index_r, :notice => 'ライセンスサーバとの通信に失敗しました。' and return
            end
          end
        else
          redirect_to :action => :index_r, :notice => '不正な入力です。' and return
        end
        # 復元されたホストが取得できなかった場合。
        unless host then
          host = SidfmdHost.where(:hostname => hostname).first
          unless host then
            redirect_to :action => :index_r, :notice => '対象ホストが見つかりません。' and return
          end
        end
        begin
          SidfmdHost.transaction do
            # カテゴリ更新
            category_names = action[:new_data][:categories]
            if category_names and category_names.size > 0 then
              categories = category_names.split(/,/).collect {|x|
                SidfmdCategory.where(:name => x).first
              }
              categories.compact!
              system_category = categories.find {|x| x.is_system }
              if system_category then
                res = host.modify_categories(system_category.id, categories.collect {|x| x.id })
                unless res then
                  raise StandardError, 'SidfmdHost#modify_categories failure'
                end
              end
            else
              SidfmdCategoriesHost.destroy_all(:sidfmd_host_id => host.id)
            end
            # アプリ更新
            product_names = action[:new_data][:products]
            if product_names and product_names.size > 0 then
              host.sidfmd_host_apps = []
              host.save
              product_names.each do |product_name|
                version = SidfmdVersion.where(:fullname => product_name).first
                if version then
                  SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                       :product_id => version.sidfmd_product.pid,
                                       :version_id => version.vid)
                else
                  product = SidfmdProduct.where(:name => product_name).first
                  if product then
                    SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                         :product_id => product.pid,
                                         :version_id => nil)
                  else
                    logger.info('ignore host app: ' + product_name)
                  end
                end
              end
            else
              SidfmdHostApp.destroy_all(:sidfmd_host_id => host.id)
            end
            # 基準日更新
            refdate = action[:new_data][:refdate]
            if refdate then
              begin
                refdate = Date.parse(refdate)
              rescue ArgumentError => e
                log_exception(e)
                logger.info('ignore invalid refdate:' + e.inspect)
                refdate = nil
              end
            end
            host.update_attribute :refdate, refdate
            # メモ更新
            memo = action[:new_data][:memo]
            if memo and memo.size > MEMO_MAX_LEN then
              redirect_to :action => :index_r, :notice => 'メモの長さが100文字を超えています。' and return
            end
            host.update_attribute :memo, memo
          end
        rescue Exception => e
          log_exception(e)
          redirect_to :action => :index_r, :notice => 'ホスト情報の更新に失敗しました。' and return
        end
      when 'remove_host' then
        # ライセンスサーバとVMでホストの削除を行う。
        action[:old_data].each do |i|
          host_id = i[:host][:id]
          hostname = i[:host][:hostname]
          sidfmdhostid = i[:host][:sidfmd_host] ? i[:host][:sidfmd_host][:id] : nil
          if host_id and hostname and sidfmdhostid then
            SidfmdHost.transaction do
              res = SidfmdHost.toggle_host_delete(host_id)
              if res then
                if res[:success] then
                  host = nil
                  if sidfmdhostid > 0 then
                    host = SidfmdHost.find(sidfmdhostid)
                  else
                    host = SidfmdHost.where(:hostname => hostname).first
                  end
                  if host then
                    host.update_attribute :flg_delete, true
                  else
                    res = SidfmdHost.establish_consistency
                  end
                else
                  redirect_to :action => :index_r, :notice => 'ライセンス数を超過するため、操作に失敗しました。' and return
                end
              else
                redirect_to :action => :index_r, :notice => 'ライセンスサーバとの通信に失敗しました。' and return
              end
            end
          end
        end
      when 'modify_host' then
        host_id = action[:new_data][:id]
        hostname = action[:new_data][:hostname]
        action[:options].each do |option|
          case option[:action]
          when 'rename_host' then
            #old_name = option[:old]
            new_name = option[:new]
            SidfmdHost.transaction do
              res = SidfmdHost.rename_host(host_id, new_name)
              if res.respond_to?(:[]) then
                if res[:success] then
                  host = SidfmdHost.where(:hostname => new_name).first
                  if host then
                    hostname = new_name
                    host.update_attribute :hostname, new_name
                  else
                    res = SidfmdHost.establish_consistency
                  end
                else
                  redirect_to :action => :index_r, :notice => '登録済のホスト名、もしくは不正な文字を指定したため、操作に失敗しました。' and return
                end
              else
                redirect_to :action => :index_r, :notice => 'ライセンスサーバとの通信に失敗しました。' and return
              end
            end
          when 'product_change' then
            # 製品変更
            host = SidfmdHost.where(:hostname => hostname).first
            if host then
              old_products = option[:old]
              new_products = option[:new]
              remove_products = old_products - new_products
              add_products = new_products - old_products
              SidfmdHostApp.transaction do
                vids = []
                pids = []
                remove_products.each do |product_name|
                  ver = SidfmdVersion.where(:fullname => product_name).first
                  if ver then
                    vids << ver.vid
                  else
                    prd = SidfmdProduct.where(:fullname => product_name).first
                    if prd then
                      pids << prd.pid
                    end
                  end
                end
                if vids.size > 0 then
                  SidfmdHostApp.where(:sidfmd_host_id => host.id).where(:version_id => vids).destroy_all
                end
                if pids.size > 0 then
                  SidfmdHostApp.where(:sidfmd_host_id => host.id).where(:product_id => pids).destroy_all
                end
                add_products.each do |product_name|
                  ver = SidfmdVersion.where(:fullname => product_name).first
                  if ver then
                    SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                         :version_id => ver.vid,
                                         :product_id => ver.sidfmd_product.pid,)
                  else
                    prd = SidfmdProduct.where(:name => product_name).first
                    if prd then
                      SidfmdHostApp.create(:sidfmd_host_id => host.id,
                                           :product_id => prd.pid)
                    end
                  end
                end
              end
            end
          when 'category_change' then
            # カテゴリ変更
            host = SidfmdHost.where(:hostname => hostname).first
            if host then
              old_categories = option[:old]
              new_categories = option[:new]
              remove_categories = old_categories - new_categories
              add_categories = new_categories - old_categories
              old_s_cat = SidfmdCategory.system.where(:name => old_categories).first
              new_s_cat = SidfmdCategory.system.where(:name => new_categories).first
              # システムカテゴリが設定されない場合は、カテゴリ変更自体を無視する。
              if (old_s_cat and new_s_cat) or
                  (!old_s_cat and new_s_cat) then
                SidfmdCategoriesHost.transaction do
                  cids = []
                  remove_categories.each do |category_name|
                    cat = SidfmdCategory.where(:name => category_name).first
                    if cat then
                      cids << cat.id
                    end
                  end
                  if cids.size > 0 then
                    SidfmdCategoriesHost.where(:sidfmd_host_id => host.id,
                                               :sidfmd_category_id => cids).destroy_all
                  end
                  add_categories.each do |category_name|
                    cat = SidfmdCategory.where(:name => category_name).first
                    if cat then
                      SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                                  :sidfmd_category_id => cat.id)
                    end
                  end
                end
              else
                logger.info('ignore change categories: old_s_cat=' + old_s_cat.inspect + ', new_s_cat=' + new_s_cat.inspect)
              end
            end
          when 'refdate_change' then
            # 基準日変更
            host = SidfmdHost.where(:hostname => hostname).first
            if host then
              begin
                SidfmdHost.transaction do
                  if option[:new] and option[:new].size > 0 then
                    host.update_attribute :refdate, Date.parse(option[:new])
                  else
                    host.update_attribute :refdate, nil
                  end
                end
              rescue ArgumentError => e
                log_exception(e)
                logger.info('ignore change refdate: new_refdate=' + option[:new].inspect)
              end
            end
          when 'memo_change' then
            # メモ変更
            host = SidfmdHost.where(:hostname => hostname).first
            if host then
              SidfmdHost.transaction do
                if option[:new] and option[:new].size > 0 then
                  if option[:new].size > MEMO_MAX_LEN then
                    logger.info("ignore change memo(over length): new_memo=#{option[:new].inspect}")
                  else
                    host.update_attribute :memo, option[:new]
                  end
                else
                  host.update_attribute :memo, nil
                end
              end
            end
          end
        end
      end
    end
    Rails.cache.clear
    redirect_to :action => :index
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  # ホスト一覧のエクスポートを行う。
  def export
    hosts = nil
    begin
      hosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => 'ライセンスサーバへのアクセスに失敗しました。'
      return
    end
    unless hosts then
      redirect_to :action => :index_r, :notice => 'データ取得に失敗しました。'
      return
    end
    header = nil
    data = nil
    if hosts and hosts[:success] then
      header = [
                'ホストID'.encode('UTF-16LE'),
                'トークン'.encode('UTF-16LE'),
                'ホスト名'.encode('UTF-16LE'),
                'カテゴリ'.encode('UTF-16LE'),
                '基準日'.encode('UTF-16LE'),
                '削除'.encode('UTF-16LE'),
                'Agentフラグ'.encode('UTF-16LE'),
                'メモ'.encode('UTF-16LE'),
                '製品名'.encode('UTF-16LE'),
               ]
      data = hosts[:hosts]
    else
      redirect_to :action => :index_r, :notice => 'データ取得に失敗しました。'
      return
    end
    if header.nil? or data.nil? then
      redirect_to :action => :index_r, :notice => 'データ取得に失敗しました。'
      return
    end
    csv = CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
      o << header
      data.each do |host|
        o << item2csv(host)
      end
    end
    send_data(csv, :filename => 'vm-hosts.csv', :disposition => 'attachment')
  end

  ### 以下、内部API

  # アプリケーション名の一部からアプリケーション情報を返す。
  def api_substr_search
    # params[keyword]
    f_params = params.fetch(:params, {}).permit(:keyword)
    keyword = f_params[:keyword]
    ret = {}
    if keyword and keyword.size > 1 then
      products = SidfmdProduct
        .where('LOWER(name) LIKE ? OR LOWER(name) LIKE ?',
               keyword.downcase + '%',
               '% ' + keyword.downcase + '%')
        .order(:category)
      products.each do |product|
        ret[product.pid] = { 'product' => product }
      end
      versions = SidfmdVersion
        .where.not(:version => 'ALL')
        .where('LOWER(fullname) LIKE ? OR LOWER(fullname) LIKE ?',
               keyword.downcase + '%',
               '% ' + keyword.downcase + '%')
        .order(:position => :desc)
      versions.each do |version|
        product = version.sidfmd_product
        if !ret.has_key?(product.pid) then
          ret[product.pid] = { 'product' => product }
        end
        ret[product.pid]['versions'] = [] unless ret[product.pid].has_key?('versions')
        ret[product.pid]['versions'] << version
      end
    end
    render :layout => false, :json => ret.to_json
  end

  # hostnameに関連づけられたアプリケーション情報を返す
  def api_get_apps_by_hostname
    f_params = params.fetch(:params, {}).permit(:hostname)
    hostname = f_params[:hostname]
    logger.info("hostname=#{hostname}")
    ret = {}
    host = SidfmdHost.where(:hostname => hostname).first
    logger.info("host=" + host.inspect)
    if host then
      apps = host.sidfmd_host_apps
      if apps then
        apps.each do |app|
          ret[app.product_id] = [] unless ret.has_key?(app.product_id)
          pid = app.product_id
          vid = app.version_id
          if !(ret[app.product_id].find {|x| x[:vid] == vid }) then
            name = ''
            fullname = ''
            if pid > 0 then
              product = SidfmdProduct.where(:pid => pid).first
              if product then
                name = product.name
              end
            end
            if vid > 0 then
              version = SidfmdVersion.where(:vid => vid).first
              if version then
                fullname = version.fullname
              end
            end
            ret[app.product_id] << {
              :pid => pid, :vid => vid,
              :name => name, :fullname => fullname, }
          end
        end
      end
    end
    render :layout => false, :json => ret.to_json
  end
 
  #新規ホスト登録時のvalidation
  def api_validate_new_host
    all_hostnames = SidfmdHost.where(:flg_delete => false).map{|host| host.hostname}
    render :layout => false, :json => all_hostnames.to_json
  end
 

  # オフラインデータの取り込み
  def load_offline_data
    f_params = params.permit(:file_data, :fileId, :initialPreview, :initialPreviewConfig, :initialPreviewThumbTags, :system_category_id, :custom_category_id)
    file_data = f_params[:file_data]
    #fileId = f_params[:fileId]
    #initialPreview = f_params[:initialPreview]
    #initialPreviewConfig = f_params[:initialPreviewConfig]
    #initialPreviewThumbTags = f_params[:initialPreviewThumbTags]
    system_category_id = (f_params[:system_category_id] || SidfmdCategory.label2id(:public_server)).to_i
    custom_category_id = (f_params[:custom_category_id] || '0').to_i
    ret = {}
    hostname = nil
    host = nil
    agent_type = :unix
    # 利用するパラメータのチェック
    begin
      if file_data then
        file_data = JSON.parse((file_data.read || '{}'), :symbolize_names => true)
        hostname = file_data[:hostname]
        unless hostname then
          logger.error('LackHostname')
          ret[:error] = 'ファイルが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        host = SidfmdHost.enable_hosts.where(:hostname => hostname).first
        if host and !@role_admin and !@user_host_ids.include?(host.id) then
          logger.error('RejectByRole')
          ret[:error] = 'アクセス権限がありません'
          render :layout => false, :json => ret.to_json and return
        end
        if file_data.has_key?(:kb) then
          agent_type = :windows
        end
      else
        logger.error('InvalidParameter')
        ret[:error] = 'ファイルが不正です'
        render :layout => false, :json => ret.to_json and return
      end
    rescue JSON::ParserError => e
      log_exception(e)
      ret[:error] = 'ファイルが不正です'
      render :lauout => false, :json => ret.to_json and return
    rescue Exception => e
      log_exception(e)
      ret[:error] = 'ファイルが不正です'
      render :layout => false, :json => ret.to_json and return
    end
    unless host then
      if @sidfm_roles.include?(:admin) then
        system_category = SidfmdCategory.find(system_category_id)
        if system_category.nil? or !system_category.is_system then
          logger.error('InvalidSystemCategoryId')
          ret[:error] = 'システムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        if custom_category_id > 0 then
          custom_category = SidfmdCategory.find(custom_category_id)
          if custom_category.nil? or custom_category.is_system then
            logger.error('InvalidCustomCategoryId')
            ret[:error] = 'カスタムカテゴリが不正です'
            render :layout => false, :json => ret.to_json and return
          end
        end
      elsif @sidfm_roles.include?(:use) then
        system_category = SidfmdCategory.find(system_category_id)
        if system_category.nil? or !system_category.is_system then
          logger.error('InvalidSystemCategoryId')
          ret[:error] = 'システムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        if custom_category_id > 0 then
          custom_category = SidfmdCategory.find(custom_category_id)
          if custom_category.nil? or custom_category.is_system then
            logger.error('InvalidCustomCategoryId')
            ret[:error] = 'カスタムカテゴリが不正です'
            render :layout => false, :json => ret.to_json and return
          end
          unless @user_category_ids.include?(custom_category_id) then
            logger.error('RejectByRole')
            ret[:error] = 'アクセス権限がありません'
            render :layout => false, :json => ret.to_json and return
          end
        else
          logger.error('InvalidCustomCategoryId')
          ret[:error] = 'カスタムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
      else
        logger.error('RejectByRole')
        ret[:error] = 'アクセス権限がありません'
        render :layout => false, :json => ret.to_json and return
      end
    end
    unless host then
      # ホスト登録処理
      # 登録APIキーのチェックは無視。
      host_apikey = SidfmdHost.register_host(hostname, nil, true, false)
      if host_apikey then
        SidfmdHost.transaction do
          host = SidfmdHost.where(:hostname => hostname).first
          unless host then
            logger.error('FailureRegistHost')
            ret[:error] = 'ホストの登録に失敗しました'
            render :layout => false, :json => ret.to_json and return
          end
          host.lock!
          host.update_attribute :flg_delete, false if host.flg_delete
          SidfmdCategoriesHost.where(:sidfmd_host_id => host.id).delete_all
          SidfmdCategoriesHost.create(:sidfmd_category_id => system_category_id,
                                      :sidfmd_host_id => host.id)
          if custom_category_id > 0 then
            SidfmdCategoriesHost.create(:sidfmd_category_id => custom_category_id,
                                        :sidfmd_host_id => host.id)
          end
        end
        Rails.cache.clear
      else
        logger.error('FailureRegistHost')
        ret[:error] = 'ホストの登録に失敗しました'
        render :layout => false, :json => ret.to_json and return
      end
    end
    # データ登録処理
    if agent_type == :windows then
      begin
        SidfmdHost.transaction do
          host.lock!
          host.update_attribute :is_agent, true unless host.is_agent
          host.create_queue_win(file_data)
        end
      rescue Exception => e
        log_exception(e)
        ret[:error] = e.message
        render :layout => false, :json => ret.to_json and return
      end
    elsif agent_type == :unix then
      begin
        SidfmdHost.transaction do
          host.lock!
          host.update_attribute :is_agent, true unless host.is_agent
          host.create_queue(file_data)
        end
      rescue Exception => e
        log_exception(e)
        ret[:error] = e.message
        render :layout => false, :json => ret.to_json and return
      end
    else
      logger.error('LastResort')
      ret[:error] = 'システムの不具合'
      render :layout => false, :json => ret.to_json and return
    end
    render :layout => false, :json => ret.to_json
  end

  def set_last_checked_at 
    f_params = params.fetch(:form, {}).permit(:time_with_hostid)
    begin 
      SidfmdHost.transaction do 
        data = JSON.parse(f_params[:time_with_hostid])
        hostid = data['hostid'].to_i
        time = data['time'].to_i
        host = SidfmdHost.find(hostid)
        if time == -1 then
          set_time = nil
        else
          if time < 10 then
            time = '0' + time.to_s
          end
          last_time = host.last_checked_at
          if last_time then
            last_time = last_time.to_time
            set_time = (last_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
            if set_time - last_time < 0 then
              set_time = set_time + 24*60*60
            end
            check_time = set_time - Time.current.to_time
            if check_time > 0 && (check_time / 60) / 60 >= 24 then
              set_time = set_time - 24*60*60
            end
          else
            set_time = (Date.today.yesterday.to_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
            if ((Time.current.to_time - set_time) / 60) / 60 >= 24 then
              set_time = set_time + 24*60*60
            end
          end 
        end
        host.update_attribute :last_checked_at, set_time
      end
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => '同期時間の設定に失敗しました。' and return
    end
    Rails.cache.clear
    redirect_to :action => :index_r and return
  end

  #同期時間の確認
  def api_get_new_last_checked_at
    f_params = params.permit(:time_with_hostid)
      data = JSON.parse(f_params[:time_with_hostid])
      hostid = data['hostid'].to_i
      time = data['time'].to_i
      host = SidfmdHost.find(hostid)
      if time == -1 then
        set_time = 'now'
        synchro_time = 'now'
      else
        if time < 10 then
          time = '0' + time.to_s
        end
        last_time = host.last_checked_at
        if last_time then
          last_time = last_time.to_time
          set_time = (last_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
          if set_time - last_time < 0 then
            set_time = set_time + 24*60*60
          end
          check_time = set_time - Time.current.to_time
          if check_time > 0 && (check_time / 60) / 60 >= 24 then
            set_time = set_time - 24*60*60
          end
        else
          set_time = (Date.today.yesterday.to_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
          if ((Time.current.to_time - set_time) / 60) / 60 >= 24 then
            set_time = set_time + 24*60*60
          end
        end
        hour = set_time.hour.to_s
        synchro_time = (set_time + 24*60*60).strftime("%Y年%m月%d日 " + hour + "時")
        set_time = set_time.strftime("%Y年%m月%d日 " + hour + "時")
      end
    render :layout => false, :json => [set_time, synchro_time].to_json
  end

end
