#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'sidfmdlm_api'

# ライセンスサーバとローカルの情報を同期させる関係上、ライセンスサーバ
# とのやりとりは全てこのモデルで行っている。
# 他の方法でAPIにアクセスしないこと。

class SidfmdHost < ActiveRecord::Base

  has_many :sidfmd_host_apps, dependent: :destroy
  has_many :sidfmd_queues, dependent: :destroy
  has_many :sidfmd_categories_hosts, dependent: :destroy
  has_and_belongs_to_many :sidfmd_categories
  has_many :sidfmd_issues, dependent: :destroy
  has_many :sidfmd_rule_t_vulsyncs, dependent: :destroy
  has_many :sidfmd_rule_cm_host_subs, dependent: :destroy

  validates :hostname, uniqueness: true

  scope :enable_hosts, -> { where(:flg_delete => false) }
  scope :categorized, -> { where(:id => SidfmdCategoriesHost.select(:sidfmd_host_id).distinct(:sidfmd_host_id)) }
  scope :not_categorized, -> { where.not(:id => SidfmdCategoriesHost.select(:sidfmd_host_id).distinct(:sidfmd_host_id)) }

  scope :select_all, -> { select(Arel.star) }
  scope :category_num, -> {
    sh = SidfmdHost.arel_table
    sch = SidfmdCategoriesHost.arel_table
    select(sch.project(sch[:sidfmd_category_id].count).where(sch[:sidfmd_host_id].eq(sh[:id])).as('category_num')).order('category_num')
  }

  scope :full_content, -> { except(:offset, :limit) }
  scope :total_size, -> { except(:order, :offset, :limit).length }

  scope :order_param, ->(p) {
    case p
    when 'id_asc' then order(:id => :asc)
    when 'id_desc' then order(:id => :desc)
    when 'hostname_asc' then order(:hostname => :asc).order(:id => :asc)
    when 'hostname_desc' then order(:hostname => :desc).order(:id => :asc)
    when 'created_at_asc' then order(:created_at => :asc).order(:id => :asc)
    when 'created_at_desc' then order(:created_at => :desc).order(:id => :asc)
    else order(:id => :asc)
    end
  }

  # SIDfmdlmAPIのライセンスAPIKEYを取得し、Sidfmに保存する。
  # 成功:true, 失敗:false
  def self.auth_api(license_key, license_password)
    ret = false
    if license_key and license_password then
      api = SIDfm::SIDfmdlmAPI.new(Sidfm.vm_config.lmapi_url, license_key)
      res = api.auth(license_password)
      if res.respond_to?(:[]) and res[:success] then
        license_apikey = res[:license_apikey]
        if license_apikey then
          Sidfm.sidfmdlm_license_key = license_key
          Sidfm.sidfmdlm_license_apikey = license_apikey
          ret = true
        end
      end
      self.establish_consistency
    end
    ret
  end

  # host_apikeyで認証を行い、成功した場合にはインスタンスを返す。
  # 失敗はnil
  def self.auth_apikey(hostname, host_apikey)
    self.where(:hostname => hostname, :host_apikey => host_apikey).first
  end

  # SIDfmdlmAPIを取得する
  # 成功:!nil, 失敗:nil
  def self.open_api
    ret = nil
    license_key = Sidfm.sidfmdlm_license_key
    license_apikey = Sidfm.sidfmdlm_license_apikey
    if license_key and license_apikey then
      ret = SIDfm::SIDfmdlmAPI.new(Sidfm.vm_config.lmapi_url, license_key, license_apikey)
    end
    ret
  end

  # 通信できず: nil
  # ライセンス超過: { :success => false }
  # 成功: {:success => true }
  def self.toggle_host_delete(host_id)
    ret = nil
    api = self.open_api
    res = api.toggle_host_delete(host_id.to_i)
    if res.respond_to?(:[]) then
      ret = res
    end
    ret
  end

  # ホスト名を変更する
  def self.rename_host(host_id, hostname)
    ret = nil
    api = self.open_api
    res = api.rename_host(host_id.to_i, hostname.to_s)
    if res.respond_to?(:[]) then
      ret = res
    end
    ret
  end

  # ライセンスマネージャから登録済ホストを取得する。
  def self.get_hosts_by_lm(opts = {})
    ret = nil
    api = self.open_api
    if api then
      ret = api.get_hosts(opts)
    end
    if ret.respond_to?(:[]) and ret[:hosts] then
      deleted = 0
      for i in 0..(ret[:hosts].size - 1) do
        host = ret[:hosts][i]
        deleted += 1 if host[:flg_delete]
        sidfmd_host = SidfmdHost.where(:hostname => host[:hostname])
        if sidfmd_host.count > 0 then
          host[:sidfmd_host] = sidfmd_host.first
        end
      end
      ret[:deleted_license_number] = deleted
    end
    ret
  end


  # lmとローカルのホストの整合性を取る。
  # ret[:success] 成否
  def self.establish_consistency
    ret = { :success => true }
    begin
      lm_hosts = self.get_hosts_by_lm
      if lm_hosts and lm_hosts[:hosts] and lm_hosts[:hosts].size > 0 then
        # lmにあってlocalに無いものをlocalに登録
        lm_hosts[:hosts].each do |host|
          SidfmdHost.transaction do
            flg_update = false
            o = self.lock.find_or_create_by(:hostname => host[:hostname]) do |i|
              i.host_apikey = host[:apikey]
              i.flg_delete = host[:flg_delete]
              flg_update = true
            end
            # find_or_create_byがfindだった場合を想定し、apikeyの不整合の
            # チェックと修正を行う。
            if host[:apikey] != o.host_apikey then
              #o.update_attribute :host_apikey, host[:apikey]
              o.host_apikey = host[:apikey]
              flg_update = true
            end
            # flg_deleteを反映
            if host[:flg_delete] != o.flg_delete then
              #o.update_attribute :flg_delete, host[:flg_delete]
              o.flg_delete = host[:flg_delete]
              flg_update = true
            end
            o.save if flg_update
          end
        end
      end
    rescue Exception => e
      ret[:success] = false
      ret[:error] = e
    end
    ret
  end

  # ライセンスAPIキーを更新する。
  # 更新に成功した場合、応答として返るAPIキーの値をデータベース
  # に保存する。
  def self.license_apikey_update
    ret = { :success => false }
    api = self.open_api
    if api then
      ret = api.update_apikey
      if ret.respond_to?(:[]) and ret[:success] then
        license_apikey = ret[:license_apikey]
        Sidfm.sidfmdlm_license_apikey = license_apikey
      end
    end
    ret
  end

  # ホストを登録し、ホストAPIキーを取得する。
  # 削除済のホストに同一ホスト名のものが存在していた場合、削除を復元して
  # そのホストAPIキーを返る。
  def self.register_host(hostname, register_apikey, agent = false, apikey_check = true)
    ret = nil
    if apikey_check and !register_apikey.nil? and
        !Sidfm.check_sidfmdlm_register_apikey(register_apikey) then
      # Register API Key不整合
      return ret
    end
    l_host = self.enable_hosts.where(:hostname => hostname).first
    if l_host then
      # localに有効なホストが登録済かつ有効であれば、ライセンスサーバへの
      # 問い合わせはしない。
      ret = l_host.host_apikey
    else
      api = self.open_api
      if api then
        SidfmdHost.transaction do
          res = api.create_host(hostname)
          logger.info(res)
          if res.respond_to?(:[]) and res[:success] and res[:host] then
            ret = res[:host][:apikey]
            l_host = self.where(:hostname => hostname).first
            if l_host then
              l_host.update_attribute :host_apikey, res[:host][:apikey]
              l_host.update_attribute :flg_delete, res[:host][:flg_delete]
              l_host.update_attribute :is_agent, agent
            else
              l_host = self.create(:hostname => res[:host][:hostname],
                                   :host_apikey => res[:host][:apikey],
                                   :flg_delete => res[:host][:flg_delete],
                                   :is_agent => agent)
              if l_host then
                SidfmdCategoriesHost.create(:sidfmd_host_id => l_host.id,
                                            :sidfmd_category_id => SidfmdCategory.system.first.id)
              end
            end
          end
        end
      end
      ret = l_host.host_apikey
    end
    ret
  end


  # queueの登録処理(agent用)
  # paramsを指定しなければ、ローカルに設定したアプリのみを登録する（agent以外用）
  def create_queue(params = nil)
    ret = SidfmdQueue.create(:sidfmd_host => self)
    if ret then
      if params.respond_to?(:[]) then
        if params[:os] then
          SidfmdQueueDist.create(:sidfmd_queue => ret,
                                 :ident => params[:os][:id],
                                 :version => params[:os][:version],
                                 :name => params[:os][:os])
        end
        if params[:packages] and params[:packages].size > 0 then
          params[:packages].each do |pkg|
            SidfmdQueuePkg.create(:sidfmd_queue => ret, :pkg => pkg)
          end
        end
        if params[:applications] and params[:applications].size > 0 then
          params[:applications].each do |app|
            if app.size == 3 then
              SidfmdQueueApp.create(:sidfmd_queue => ret,
                                    :name => app[0],
                                    :version => app[1],
                                    :path => app[2])
            end
          end
        end
      end
      if self.sidfmd_host_apps.size > 0 then
        self.sidfmd_host_apps.each do |app|
          SidfmdQueueLapp.create(:sidfmd_queue => ret,
                                 :product_id => app.product_id,
                                 :version_id => app.version_id)
        end
      end
      ret.update_digest
    end
    ret
  end

  # queueの登録処理(windows用)
  # paramsを指定しなければ、ローカルに設定したアプリのみを登録する(agent以外用)
  def create_queue_win(params = nil)
    ret = SidfmdQueue.create(:sidfmd_host => self)
    if ret then
      if params.respond_to?(:[]) then
        if params[:os] and params[:kb] then
          osname = nil
          if params[:osname] then
            temp_osname = params[:osname].split(/\|/)
            if temp_osname.size > 0 then
              osname = temp_osname.first
            end
          end
          temp_id = SidfmdQueueWin.select(:id).max
          temp_id = temp_id ? temp_id.id + 1 : 1
          winqueue = SidfmdQueueWin.create(:id => temp_id,
                                           :sidfmd_queue => ret,
                                           :last_date => params[:last_date],
                                           :os => params[:os],
                                           :osname => osname)
          if winqueue and params[:kb].kind_of?(Enumerable) then
            params[:kb].each do |kb|
              temp_id = SidfmdQueueWinDetail.select(:id).max
              temp_id = temp_id ? temp_id.id + 1 : 1
              kbqueue = SidfmdQueueWinDetail.create(:id => temp_id,
                                                    :sidfmd_queue_win => winqueue,
                                                    :title => kb[:title],
                                                    :kb => kb[:id],
                                                    :severity => kb[:severity])
            end
          end
        end
      end
      if self.sidfmd_host_apps.size > 0 then
        self.sidfmd_host_apps.each do |app|
          SidfmdQueueLapp.create(:sidfmd_queue => ret,
                                 :product_id => app.product_id,
                                 :version_id => app.version_id)
        end
      end
      ret.update_digest
    end
    ret
  end

  # 最新のqueueを返す。
  def get_last_queue
    sidfmd_queues.order(:created_at => :desc).limit(1).first
  end

  # 最新の未確認queueを返す。
  # 具体的には以下の動作を行う。
  # 1. 最新のqueueを取得
  # 2. done = false であれば1.のqueueを返す
  # 3. done = true であれば、nilを返す。
  # 最新ではないqueueにdone=falseのものがあってもqueueを返さないことに注意。
  def get_unprocessed_queue
    queue = sidfmd_queues.order(:created_at => :desc).limit(1).first
    queue ? (queue.done ? nil : queue) : nil
  end

  # 最新の確認済のqueueを返す
  def get_last_processed_queue
    queue = sidfmd_queues.order(:created_at => :desc).limit(1).first
    queue ? (queue.done ? queue : nil) : nil
  end

  # 最新の確認済かつissue更新処理未実施のqueueを返す。
  def get_last_unchecked_queue 
    queue = sidfmd_queues.order(:created_at => :desc).limit(1).first
    queue ? ((queue.done && !queue.checked) ? queue : nil) : nil
  end

  # システムカテゴリが設定されていればtrueを返す。
  def has_system_category?
    SidfmdCategoriesHost.where(:sidfmd_host_id => id).where(:sidfmd_category_id => SidfmdCategory.system.select(:id).to_a).count > 0
  end

  # システムカテゴリを返す。
  def system_category
    ret = sidfmd_categories_hosts.where(:sidfmd_category_id => SidfmdCategory.system.select(:id).to_a).first
    ret ? ret.sidfmd_category : nil
  end

  # システムカテゴリ以外に登録されているカテゴリを返す。
  def no_system_categories(user = nil)
    ret = sidfmd_categories_hosts.where.not(:sidfmd_category_id => SidfmdCategory.system.select(:id).to_a)
    ret = ret.where(:sidfmd_category_id => SidfmdCategoriesUser.where(:user_id => user.id).select(:sidfmd_category_id)) if user
    ret
  end

  # カテゴリを更新する。
  # scid: system category
  # cids: other categories
  def modify_categories(scid, cids, l_cat_ids = nil)
    unless l_cat_ids then
      cats = []
      cats << SidfmdCategory.find(scid)
      cats += cids.collect {|x| c = SidfmdCategory.find(x) ; c.is_system ? nil : c }
      self.sidfmd_categories = cats.compact.uniq
      self.save
    else
      prev_cat_ids = self.sidfmd_categories.where(:is_system => false).collect {|x| x.id }
      mod_cat_ids = prev_cat_ids - l_cat_ids
      cats = mod_cat_ids.collect {|x| SidfmdCategory.find(x) }
      cats << SidfmdCategory.find(scid)
      cats += cids.collect {|x| c = SidfmdCategory.find(x) ; c.is_system ? nil : c }
      self.sidfmd_categories = cats.compact.uniq
      self.save
    end
  end

  # agentの場合のみ、OS(Distribution)の名前とバージョンを示す文字列を返す。
  def get_dist
    ret = '（不明）'
    q = sidfmd_queues.last
    if q then
      w = q.sidfmd_queue_win
      if w then
        o = w.os_info
        if o then
          ret = o[:name]
        end
      else
        d = q.sidfmd_queue_dist
        if d then
          ret = d.name + ' ' + d.version
        end
      end
    end
    ret
  end

  # agentの場合のみ、OS情報（SidfmdQueueDist）を返す。
  def get_dist_data
    ret = nil
    q = sidfmd_queues.last
    if q then
      w = q.sidfmd_queue_win
      unless w then
        d = q.sidfmd_queue_dist
        if d then
          ret = d
        end
      end
    end
    ret
  end

  # agentの場合はSidfmdQueueDistから取得したOS情報、
  # 手動入力の場合はSidfmdHostAppのcategory=OSの情報があればそれをOSとして
  # 返す。
  # 適合するものが何もなければ nil を返す。
  # { :osname, :dist, :app }
  # caution: :distがあれば:app はありません。
  def get_os
    ret = {}
    dist = get_dist_data
    if dist then
      ret[:dist] = dist
      ret[:osname] = "#{dist.name} #{dist.version}"
      if 'Ubuntu' == dist.name then
        ret[:osname] += ' LTS'
      end
    else
      q = self.sidfmd_queues.last
      if q then
        w = q.sidfmd_queue_win
        if w then
          o = w.os_info
          if o[:version_id] > 0 then
            prod = SidfmdProduct.where(:pid => o[:product_id]).first
            ver = SidfmdVersion.where(:vid => o[:version_id]).first
            ret[:osname] = o[:name]
            ret[:app] = [ { :osname => o[:name], :product => prod, :version => ver } ]
            return ret
          end
        end
      end
      apps = self.sidfmd_host_apps
      if apps and apps.count > 0 then
        os = []
        apps.each do |app|
          prod = nil
          ver = nil
          if app.version_id and app.version_id > 0 then
            ver = SidfmdVersion.where(:vid => app.version_id).first
            if ver then
              prod = ver.sidfmd_product
            end
          end
          if prod.nil? and app.product_id  and app.product_id > 0 then
            prod = SidfmdProduct.where(:pid => app.product_id).first
          end
          if prod and prod.category == 'OS' then
            os << { :product => prod, :version => ver }
          end
        end
        prodver = os.first
        if prodver then
          ret[:app] = prodver
          if prodver[:version] then
            ret[:osname] = prodver[:version].fullname
          else
            ret[:osname] = prodver[:product].name
          end
        end
      else
        ret = nil
      end
    end
    ret
  end

  # 最大SRI値を返す
  # STATUS_UNFINISHEDのみ。
  def get_max_sri
    sris = nil
    s_cat = system_category
    if s_cat then
      sris = SidfmdIssueSri.joins(:sidfmd_issue).joins(:sidfmd_issue => [ :issue ]).where(:sidfmd_issues => { :sidfmd_host_id => id }).where(:category_id => s_cat.id).where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
    end
    sris ? sris.minimum(:risk_id) : nil
  end


  # 最大CVSSを返す。
  # STATUS_UNFINISHEDのみ（未対処）で、c3_base_score, c2_base_score
  # 区別せずに最大値を返す。
  def get_max_cvss
    base = SidfmdIssueCve.joins(:sidfmd_issue).joins(:sidfmd_issue => [ :issue ]).where(:sidfmd_issues => { :sidfmd_host_id => id }).where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
    c3 = base.maximum(:c3_base_score)
    c3 = 0.0 unless c3
    c2 = base.maximum(:c2_base_score)
    c2 = 0.0 unless c2
    c3 > c2 ? c3 : c2
  end

  # 最大Risk,Areaを返す。
  # { :risk => SidfmdRisk, :area => SidfmdArea ]
  # STATUS_UNFINISHEDのみ（未対処）
  # area は、remote > local > client で比較する。
  def get_max_softek
    base = sidfmd_issues.joins(:issue).where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
    risk_id = base.maximum(:risk_id)
    area_id = 0
    if base.where(:area_id => 3).count > 0 then
      area_id = 3
    elsif base.where(:area_id => 1).count > 0 then
      area_id = 1
    else
      area_id = 2
    end
    { :risk => SidfmdRisk.where(:risk_id => risk_id).first,
      :area => SidfmdArea.where(:area_id => area_id).first }
  end

  # issueの状態をカウントして返す。
  # { status_id => issueの件数, ... }の形となる。
  def status_count
    status2color = {
      SidfmdIssue::STATUS_NOTAFFECT => :status_notaffect,
      SidfmdIssue::STATUS_FIXED => :status_fixed,
      SidfmdIssue::STATUS_ONGOING => :status_ongoing,
      SidfmdIssue::STATUS_NOTCONFIRMED => :status_notconfirmed
    }
    cnt = SidfmdIssue.joins(:issue).where(:sidfmd_host_id => id).group(:status_id).order('issues.status_id').count
    ret = []
    cnt.each_pair do |k, v|
      ret << { :id => id, :value => v, :name => IssueStatus.find(k).name, :color => status2color[k], :placement => 'top' }
    end
    ret
  end

  # 対策状況
  def status_info
    cnt = SidfmdIssue.joins(:issue).where(:sidfmd_host_id => id).group(:status_id).count
    ret = {}
    ret[:sum] = cnt.inject(0) {|r, x| r += x[1]}
    ret[:finished] = SidfmdIssue::STATUS_FINISHED
    ret[:unfinished] = SidfmdIssue::STATUS_UNFINISHED
    ret[:finished_cnt] = ret[:finished].inject(0) {|r, x| r += cnt[x] || 0 }
    ret[:unfinished_cnt] = ret[:unfinished].inject(0) {|r, x| r += cnt[x] || 0 }
    if ret[:unfinished_cnt] > 0 then
      ret[:finished_percent] = (ret[:finished_cnt].to_f / ret[:unfinished_cnt].to_f * 100.0).round
    else
      if ret[:finished_cnt] > 0 then
        ret[:finished_percent] = 100.0
      else
        ret[:finished_percent] = 0.0
      end
    end
    ret[:unfinished_percent] = 100.0 - ret[:finished_percent]
    IssueStatus.all.each do |status|
      ret[status.id] = { :status => status, :cnt => cnt[status.id] || 0,
        :percent => ret[:sum] != 0 ? (cnt[status.id].to_f / ret[:sum].to_f * 100.0).round : 0  }
    end
    ret
  end

  # 対策予定日(due_date)設定済の脆弱性の件数の割合
  # STATUS_FINISHEDはカウントから除外。
  # STATUS_UNFINISHEDかつ issues.due_dateがnil以外を設定済として扱う。
  def due_date_scheduled_percent
    s = sidfmd_issues.joins(:issue).where(:issues => { :status_id => SidfmdIssue::STATUS_UNFINISHED })
    cnt_all = s.count
    cnt_due_date = s.where.not(:issues => { :due_date => nil }).count
    cnt_all > 0 ? cnt_due_date.to_f / cnt_all.to_f : nil
  end

  # 対応状況(%)を返す。
  # STATUS_FINISHEDを対応済としてカウント。
  def finished_percent
    s = sidfmd_issues.joins(:issue)
    cnt_all = s.count
    cnt_finished = s.where(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED }).count
    cnt_all > 0 ? cnt_finished.to_f / cnt_all.to_f : nil
  end

  def has_exploitable?
    SidfmdIssueCve.joins(:sidfmd_issue).where(:sidfmd_issues => { :sidfmd_host_id => id }).where('c2_exploitable IS TRUE OR c3_exploitable IS TRUE').count > 0
  end

  # 対応が必要なホストかどうかを返す
  def corresp_required?(index)
    case index
    when :sri then
      corresp_required_sri?
    when :cvss then
      corresp_required_cvss?
    when :softek then
      corresp_required_softek?
    else
      false
    end
  end

  # 対応が必要なホストかどうかを返す(基準:SRI)。
  def corresp_required_sri?
    ret = nil

    cr_sri = Sidfm.corresp_required_sri
    cr_month = Sidfm.corresp_required_month

    s_cat = system_category
    if s_cat then
      ret = SidfmdIssueSri.joins(:sidfmd_issue).joins(:sidfmd_issue => [ :issue ]).where(:sidfmd_issues => { :sidfmd_host_id => id }).where(:category_id => s_cat.id).where('sidfmd_issue_sris.risk_id <= ?', cr_sri)
      if cr_month then
        ret = ret.where('issues.created_on <= ?', cr_month.month.ago)
      end
    end

    ret ? ret.count > 0 : false
  end

  # 対応が必要なホストかどうかを返す(基準:CVSS)。
  def corresp_required_cvss?
    cr_cvss = Sidfm.corresp_required_cvss
    cr_month = Sidfm.corresp_required_month
    ret = SidfmdIssueCve.joins(:sidfmd_issue).joins(:sidfmd_issue => [ :issue ]).where(:sidfmd_issues => { :sidfmd_host_id => id }).where('c3_base_score > ? OR c2_base_score > ?', cr_cvss, cr_cvss)
    if cr_month > 0 then
      ret = ret.where('issues.created_on <= ?', cr_month.month.ago)
    end
    ret ? ret.count > 0 : false
  end

  # 対応が必要なホストかどうかを返す(基準:Softek)。
  # なお、cr_areaの値は以下のように解釈される。
  #    area 3 - remote only
  #    area 1 - remote,local only
  #    area 2 - all
  def corresp_required_softek?
    cr_risk = Sidfm.corresp_required_risk
    cr_risk = [ 3, 1 ] if cr_risk == 1
    cr_risk = [ 1, 2, 3 ] if cr_risk == 2
    cr_area = Sidfm.corresp_required_area
    cr_month = Sidfm.corresp_required_month

    ret = SidfmdIssue.joins(:issue).where(:sidfmd_host_id => 6).where('risk_id >= ?', cr_risk)
    if cr_area != 2 then
      ret = ret.where(:area_id => cr_area)
    end
    if cr_month then
      ret = ret.where('issues.created_on < ?', cr_month.month.ago)
    end
    ret ? ret.count > 0 : false
  end

  # 有効なホストの数を返す。
  # category_idが指定された場合は、指定されたカテゴリ内でフィルタした結果
  # を返す。
  def self.enable_host_size(category_id = 0)
    hosts = self.enable_hosts.categorized
    if category_id > 0 then
      hosts =
        hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
    end
    hosts.count
  end

  # heatmap用データを返す。
  # 危険度は、各ホストで残存している脆弱性のもっとも危険度の高いものと
  # して扱う。
  def self.enable_all_hosts(idx = :sri, category_id = 0)
    case idx
    when :sri then self.enable_all_hosts_sri(category_id)
    when :cvss then self.enable_all_hosts_cvss(category_id)
    when :softek then self.enable_all_hosts_softek(category_id)
    end
  end

  # heatmap用データを返す。
  # 危険度は、各ホストで残存している脆弱性のもっとも危険度の高いものと
  # して扱う。
  def self.enable_all_hosts_sri(category_id = 0)
    hosts = self.enable_hosts.categorized
    if category_id > 0 then
      hosts =
        hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
    end
    hosts
      .select(:id)
      .select(:hostname)
      .select('(SELECT MAX(risk_id) FROM sidfmd_issues WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id) AS risk_id')
      .select('(SELECT name FROM sidfmd_risks WHERE id=(SELECT MAX(risk_id) FROM sidfmd_issues WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id)) AS risk_name')
  end

  def self.enable_all_hosts_cvss(category_id = 0)
    hosts = SidfmdHost
      .joins('INNER JOIN (' +
             SidfmdIssue
               .joins(:sidfmd_host)
               .joins(:sidfmd_issue_cves)
               .select(:sidfmd_host_id)
               .select('COALESCE(MAX(c3_base_score),MAX(c2_base_score)) AS score')
               .group(:sidfmd_host_id).to_sql +
             ') AS scores ON scores.sidfmd_host_id=sidfmd_hosts.id')
      .where(:flg_delete => false)
    if category_id > 0 then
      hosts = hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
    else
      hosts = hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => {
                 :sidfmd_category_id =>
                 SidfmdCategory.where(:is_system => true).select(:id)
               })
    end
    hosts
      .select(:id)
      .select(:hostname)
      .select('scores.score AS risk_id')
      .select('TO_CHAR(scores.score,\'99D9\') AS risk_name')
  end

  def self.enable_all_hosts_softek(category_id = 0)
    hosts = SidfmdHost
      .joins('INNER JOIN (' +
             SidfmdIssue
               .select(:sidfmd_host_id)
               .select('MAX(risk_id) AS risk_max')
               .group(:sidfmd_host_id).to_sql +
             ') AS risks ON risks.sidfmd_host_id=sidfmd_hosts.id')
      .where(:flg_delete => false)
    if category_id > 0 then
      hosts = hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
    else
      hosts = hosts
        .joins(:sidfmd_categories_hosts)
        .where(:sidfmd_categories_hosts => {
                 :sidfmd_category_id =>
                 SidfmdCategory.where(:is_system => true).select(:id)
               })
    end
    hosts
      .select(:id)
      .select(:hostname)
      .select('risks.risk_max AS risk_id')
      .select('(SELECT name_en FROM sidfmd_risks WHERE sidfmd_risks.risk_id=risks.risk_max) AS risk_name')
  end


  # 対応状況毎のホスト数分布
  def self.distribution_status(category_id = 0)
    dists = SidfmdIssue.enable_hosts.categorized.joins(:issue)
    if category_id > 0 then
      dists = dists.joins(:sidfmd_host).joins(:sidfmd_host => [ :sidfmd_categories_hosts ]).where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
    end
    dists = dists.group(:sidfmd_host_id, :status_id).count.keys
    host_status = {}
    dists.each do |dist|
      if host_status.has_key?(dist[0]) then
        if host_status[dist[0]] < dist[1] then
          host_status[dist[0]] = dist[1]
        end
      else
        host_status[dist[0]] = dist[1]
      end
    end
    {
      :notconfirmed => host_status.inject(0) {|r, x|
        x[1] == SidfmdIssue::STATUS_NOTCONFIRMED ? r+= 1 : r
      },
      :ongoing => host_status.inject(0) {|r, x|
        x[1] == SidfmdIssue::STATUS_ONGOING ? r += 1 : r
      },
      :fixed => host_status.inject(0) {|r, x|
        x[1] == SidfmdIssue::STATUS_FIXED ? r += 1 : r
      },
      :notaffect => host_status.inject(0) {|r, x|
        x[1] == SidfmdIssue::STATUS_NOTAFFECT ? r += 1 : r
      },
    }
  end

  # 危険度毎のホスト数分布
  def self.distribution_risk(idx = :sri, category_id = 0)
    case idx
    when :sri then self.distribution_risk_sri(category_id)
    when :cvss then self.distribution_risk_cvss(category_id)
    when :softek then self.distribution_risk_softek(category_id)
    end
  end

  # SRI重大度でのホスト数分布
  def self.distribution_risk_sri(category_id = 0)
    hosts = SidfmdIssueSri
      .joins(:sidfmd_issue)
      .joins(:sidfmd_issue => [ :sidfmd_host ])
      .joins(:sidfmd_issue =>
             [ :sidfmd_host => [ :sidfmd_categories_hosts ]])
      .where(:sidfmd_hosts => { :flg_delete => false })
      .where(:sidfmd_issue_sris =>
             { :category_id  =>
               SidfmdCategoriesHost.arel_table[:sidfmd_category_id] })
    if category_id > 0 then
      hosts = hosts.where(:sidfmd_categories_hosts =>
                          { :sidfmd_category_id => category_id })
    else
      hosts = hosts.where(:sidfmd_categories_hosts =>
                          { :sidfmd_category_id =>
                            SidfmdCategory.where(:is_system => true) })
    end
    hosts = hosts
      .group(SidfmdIssue.arel_table[:sidfmd_host_id],
             SidfmdIssueSri.arel_table[:risk_id])
      .count.keys
    host_risk = {}
    hosts.each do |host|
      if host_risk.has_key?(host[0]) then
        if host_risk[host[0]] < host[1] then
          host_risk[host[0]] = host[1]
        end
      else
        host_risk[host[0]] = host[1]
      end
    end
    ret = {
      :critical => host_risk.inject(0) {|r, x| x[1] == 4 ? r += 1 : r },
      :high => host_risk.inject(0) {|r, x| x[1] == 3 ? r += 1 : r },
      :medium => host_risk.inject(0) {|r, x| x[1] == 2 ? r += 1 : r },
      :low => host_risk.inject(0) {|r, x| x[1] == 1 ? r += 1 : r },
    }
    ret_host_num = ret.inject(0) {|r, x| r += x[1] }
    ret[:none] = self.enable_host_size(category_id) - ret_host_num
    ret
 end

  # CVSS重大度でのホスト数分布
  def self.distribution_risk_cvss(category_id = 0)
    ret = {
      :notreg => {
        :high => nil,
        :low => nil,
        :cnt => 0,
      },
      :d4 => {
        :high => 10.0,
        :low => 9.0,
        :cnt => 0,
      },
      :d3 => {
        :high => 8.9,
        :low => 7.0,
        :cnt => 0,
      },
      :d2 => {
        :high => 6.9,
        :low => 4.0,
        :cnt => 0,
      },
      :d1 => {
        :high => 3.9,
        :low => 0.0,
        :cnt => 0,
      },
    }
    ret.each_pair do |k, v|
      hosts = SidfmdHost
        .joins('INNER JOIN (' +
               SidfmdIssue
                 .joins(:sidfmd_host)
                 .joins(:sidfmd_issue_cves)
                 .select(:sidfmd_host_id)
                 .select('COALESCE(MAX(c3_base_score),MAX(c2_base_score)) AS score')
                 .group(:sidfmd_host_id).to_sql +
               ') AS scores ON scores.sidfmd_host_id=sidfmd_hosts.id')
        .where(:flg_delete => false)

      case k
      when :notreg then
        hosts = hosts
          .where('score IS NULL')
      when :d3, :d2, :d1 then
        hosts = hosts
          .where('score >= ? AND score < ?', v[:low], v[:high])
      when :d4 then
        hosts = hosts
          .where('score >= ? AND score <= ?', v[:low], v[:high])
       end
      if category_id > 0 then
        hosts = hosts
          .joins(:sidfmd_categories_hosts)
          .where(:sidfmd_categories_hosts => { :sidfmd_category_id => category_id })
      else
        hosts = hosts
          .joins(:sidfmd_categories_hosts)
          .where(:sidfmd_categories_hosts => {
                   :sidfmd_category_id =>
                   SidfmdCategory.where(:is_system => true).select(:id)
                 })
      end
      v[:cnt] = hosts.count
    end

    ret
  end


  # SofTek危険度でのホスト数分布
  def self.distribution_risk_softek(category_id = 0)
    hosts = SidfmdHost
      .joins(:sidfmd_issues)
      .joins(:sidfmd_categories_hosts)
      .where(:flg_delete => false)
    if category_id > 0 then
      hosts = hosts
        .where(:sidfmd_categories_hosts =>
               { :sidfmd_category_id => category_id })
    else
      hosts = hosts
        .where(:sidfmd_categories_hosts =>
               { :sidfmd_category_id =>
                 SidfmdCategory.where(:is_system => true).select(:id) })
    end
    hosts = hosts
      .group(:id, :risk_id).count.keys
    host_risk = {}
    hosts.each do |host|
      if host_risk.has_key?(host[0]) then
        if host_risk[host[0]] < host[1] then
          host_risk[host[0]] = host[1]
        end
      else
        host_risk[host[0]] = host[1]
      end
    end
    ret = {
      :critical => host_risk.inject(0) {|r, x| x[1] == 4 ? r += 1 : r },
      :high => host_risk.inject(0) {|r, x| x[1] == 3 ? r += 1 : r },
      :medium => host_risk.inject(0) {|r, x| x[1] == 2 ? r += 1 : r },
      :low => host_risk.inject(0) {|r, x| x[1] == 1 ? r += 1 : r },
    }
    ret
  end

  # 対応が必要なホスト数（対応が必要な脆弱性が未対応で存在するホストの数）
  # を返す。
  # category_idが指定された場合は、指定されたカテゴリ内でフィルタした結果
  # を返す。
  def self.corresp_host_num(index, category_id = 0)
    case index
    when :sri then
      corresp_host_num_sri(category_id)
    when :cvss then
      corresp_host_num_cvss(category_id)
    when :softek then
      corresp_host_num_softek(category_id)
    else
      0
    end
  end

  # 対応が必要なホスト数を返す（基準：SRI）
  def self.corresp_host_num_sri(category_id = 0)
    hosts = SidfmdIssueSri
      .joins(:sidfmd_issue)
      .joins(:sidfmd_issue => [ :sidfmd_host ])
      .joins(:sidfmd_issue =>
             [ :sidfmd_host => [ :sidfmd_categories_hosts ]])
      .where(:sidfmd_hosts => { :flg_delete => false })
      .where(:sidfmd_issue_sris =>
             { :category_id  =>
               SidfmdCategoriesHost.arel_table[:sidfmd_category_id] })
    if category_id > 0 then
      hosts = hosts.where(:sidfmd_categories_hosts =>
                          { :sidfmd_category_id => category_id })
    else
      hosts = hosts.where(:sidfmd_categories_hosts =>
                          { :sidfmd_category_id =>
                            SidfmdCategory.where(:is_system => true) })
    end
    hosts = hosts
      .where(SidfmdIssueSri.arel_table[:risk_id]
               .gteq(Sidfm.corresp_required_sri))
      .group(SidfmdIssue.arel_table[:sidfmd_host_id])
    hosts.count.keys.size
  end

  # 対応が必要なホスト数を返す（基準：CVSS）
  def self.corresp_host_num_cvss(category_id = 0)
    cr_cvss = Sidfm.corresp_required_cvss
    hosts = SidfmdIssueCve
      .joins(:sidfmd_issue)
      .joins(:sidfmd_issue => [ :sidfmd_host ])
      .joins(:sidfmd_issue =>
             [ :sidfmd_host => [ :sidfmd_categories_hosts ]])
      .where(:sidfmd_hosts => { :flg_delete => false })
      .where('c3_base_score >= ? OR c2_base_score >= ?', cr_cvss, cr_cvss)
    if category_id > 0 then
      hosts =
        hosts.where(:sidfmd_categories_hosts =>
                    { :sidfmd_category_id => category_id })
    else
      hosts =
        hosts.where(:sidfmd_categories_hosts =>
                    { :sidfmd_category_id =>
                      SidfmdCategory.where(:is_system => true) })
    end
    hosts
      .select('sidfmd_issues.sidfmd_host_id')
      .distinct(:sidfmd_host_id)
      .count
  end
 
  # 対応が必要なホスト数を返す（基準：SofTek）
  def self.corresp_host_num_softek(category_id = 0)
    hosts = SidfmdIssue
      .joins(:sidfmd_host)
      .joins(:sidfmd_host => [ :sidfmd_categories_hosts ])
      .where(:sidfmd_hosts => { :flg_delete => false })
      .where('sidfmd_issues.risk_id >= ?', Sidfm.corresp_required_risk)
    if category_id > 0 then
      hosts =
        hosts.where(:sidfmd_categories_hosts =>
                    { :sidfmd_category_id => category_id })
    else
      hosts =
        hosts.where(:sidfmd_categories_hosts =>
                    { :sidfmd_category_id =>
                      SidfmdCategory.where(:is_system => true) })
    end
    hosts.group(:sidfmd_host_id).count.keys.size
  end

  # htmlidに対応するsidfmd_issuesから、ホスト一覧を取得し返す。
  def self.get_hosts_htmlid(htmlid)
    self.enable_hosts.where(:id => SidfmdIssue.enable_hosts.where(:html_id => htmlid).collect {|x| x.sidfmd_host_id }).select(:id, :hostname, :flg_delete, :is_agent, :created_at, :updated_at)
  end


  # ホスト一覧のデータを生成する。
  def self.create_all_content(**opts)
    [ :category_id, :role_admin, :corresp_required ].each do |k|
      raise ArgumentError, "must be #{k.to_s}" unless opts.has_key?(k)
    end
    unless opts[:role_admin] then
      [ :user_host_ids, :user_category_ids ].each do |k|
        raise ArgumentError, "must be #{k.to_s}" unless opts.has_key?(k)
      end
    end
    cache_key = '//hosts'
    if opts[:role_admin] then
      cache_key += "/admin/#{opts[:category_id]}"
    else
      cache_key += "/user/#{opts[:user_category_ids].sort.join(',')}/#{opts[:category_id]}"
    end
    opts[:no_cache] ? create_all_content_base(opts) : Rails.cache.fetch(cache_key, { :expires_in => 1.hour, :race_condition_ttl => 1.minute }) { create_all_content_base(opts) }
  end

  # sub method for create_all_content
  def self.create_all_content_base(opts)
    ret = []
    hosts = SidfmdHost.enable_hosts
    if opts[:host_ids] then
      hosts = hosts.where(:id => opts[:host_ids])
    end
    if opts[:role_admin] then
      if opts[:category_id] > 0 then
        hosts =
          hosts
          .where(:id =>
                 SidfmdCategoriesHost
                   .joins(:sidfmd_host)
                   .where(:sidfmd_hosts => { :flg_delete => false })
                   .where(:sidfmd_category_id => opts[:category_id])
                   .select(:sidfmd_host_id)
                   .distinct)
      else
        hosts = hosts.categorized
      end
    else
      if opts[:category_id] > 0 then
        cat = SidfmdCategory.find(opts[:category_id])
        if cat then
          hosts = hosts
            .categorized
            .where(:id => opts[:user_host_ids])
            .where(:id =>
                   SidfmdCategoriesHost
                     .joins(:sidfmd_host)
                     .where(:sidfmd_hosts => { :flg_delete => false })
                     .where(:sidfmd_category_id => opts[:category_id])
                     .select(:sidfmd_host_id)
                     .distinct)
        else
          hosts = hosts
            .categorized
            .where(:id => opts[:user_host_ids])
        end
      else
        hosts = hosts
          .categorized
          .where(:id => opts[:user_host_ids])
      end
    end

    hosts = hosts
      .select(:id)
      .select(:hostname)
      .select(:last_checked_at)
      .select("
(
 SELECT CAST(sidfmd_categories.id AS text) || ':' || sidfmd_categories.name
 FROM sidfmd_categories
 INNER JOIN sidfmd_categories_hosts
  ON sidfmd_categories_hosts.sidfmd_category_id=sidfmd_categories.id
 WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id AND
       sidfmd_categories.is_system IS TRUE
 ORDER BY sidfmd_categories.id
 LIMIT 1
) AS system_category")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT CAST(sidfmd_categories.id AS text) || ':' || sidfmd_categories.name
 FROM sidfmd_categories
 INNER JOIN sidfmd_categories_hosts
  ON sidfmd_categories_hosts.sidfmd_category_id=sidfmd_categories.id
 WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id AND
       sidfmd_categories.is_system IS FALSE
 ORDER BY sidfmd_categories.id
),',')) AS no_system_categories")
      .select("
(
 SELECT BOOL_OR(c32_exploitable)
 FROM (
   SELECT BOOL_OR(COALESCE(c3_exploitable,c2_exploitable,FALSE)) AS c32_exploitable
   FROM sidfmd_issue_cves
   INNER JOIN sidfmd_issues
    ON sidfmd_issues.id=sidfmd_issue_cves.sidfmd_issue_id
   INNER JOIN issues
    ON sidfmd_issues.issue_id=issues.id
   WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
         issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
   GROUP BY c3_exploitable,c2_exploitable
 ) AS exploitables
 LIMIT 1
) AS exploitable")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT CAST(BOOL_OR(COALESCE(c3_exploitable,c2_exploitable,FALSE)) AS text) || ':' ||
        CAST(MAX(COALESCE(c3_base_score,c2_base_score)) AS text)
 FROM sidfmd_issue_cves
 INNER JOIN sidfmd_issues
  ON sidfmd_issues.id=sidfmd_issue_cves.sidfmd_issue_id
 INNER JOIN issues
  ON sidfmd_issues.issue_id=issues.id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
 GROUP BY c3_exploitable,c2_exploitable
),',')) AS cvss")
      .select("
(
 SELECT MAX(max_sri_sub.max_score)
 FROM (
  SELECT MAX(sidfmd_issue_sris.risk_id) AS max_score
  FROM sidfmd_issues
  INNER JOIN issues
   ON sidfmd_issues.issue_id=issues.id
  INNER JOIN sidfmd_issue_sris
   ON sidfmd_issues.id=sidfmd_issue_sris.sidfmd_issue_id
  WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
        sidfmd_issues.html_id > 0 AND
        issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')}) AND
        sidfmd_issue_sris.category_id=
         (SELECT sidfmd_category_id FROM sidfmd_categories_hosts
          INNER JOIN sidfmd_categories
           ON sidfmd_categories.id=sidfmd_categories_hosts.sidfmd_category_id
          WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id AND
           sidfmd_categories.is_system IS TRUE
          ORDER BY sidfmd_categories.id
          LIMIT 1)
  UNION
  SELECT MAX(sidfmd_issues.risk_id) AS max_score
  FROM sidfmd_issues
  INNER JOIN issues
   ON sidfmd_issues.issue_id=issues.id
  WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
        sidfmd_issues.html_id < 0 AND
        issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
 ) AS max_sri_sub
 LIMIT 1
) AS max_sri")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT sidfmd_issues.risk_id
 FROM sidfmd_issues
 INNER JOIN issues
  ON sidfmd_issues.issue_id=issues.id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
 ORDER BY sidfmd_issues.id
),',')) AS risk_ids")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT sidfmd_issues.area_id
 FROM sidfmd_issues
 INNER JOIN issues
  ON sidfmd_issues.issue_id=issues.id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
 ORDER BY sidfmd_issues.id
),',')) AS area_ids")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT (CASE WHEN issues.due_date IS NULL THEN '' ELSE TO_CHAR(issues.due_date, 'YYYY/MM/DD') END)
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id
),',')) AS due_dates")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT (CASE WHEN issues.due_date IS NULL THEN '' ELSE TO_CHAR(issues.due_date, 'YYYY/MM/DD') END)
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
),',')) AS due_dates_unfinished")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT (CASE WHEN issues.due_date IS NULL
              THEN ''
              ELSE TO_CHAR(issues.created_on, 'YYYY/mm/dd HH24:MI:SS OF') END)
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id
 ORDER BY sidfmd_issues.id
),',')) AS created_ons")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT CAST(issues.id AS text) || ':' ||
        CAST(issues.status_id AS text)  || ':'||
        TO_CHAR(issues.created_on, 'YYYY/mm/dd HH24:MI:SS OF')
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id
),',')) AS tl_issue_ids")
      .select("
(
 SELECT COUNT(id)
 FROM (
  SELECT sidfmd_issues.id,MAX(sidfmd_issue_sris.risk_id) AS max_sri
  FROM sidfmd_issue_sris
  INNER JOIN sidfmd_issues ON sidfmd_issues.id=sidfmd_issue_sris.sidfmd_issue_id
  INNER JOIN issues ON issues.id=sidfmd_issues.issue_id
  WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
        issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')}) AND
        sidfmd_issue_sris.category_id=(
         SELECT sidfmd_category_id
         FROM sidfmd_categories_hosts
         INNER JOIN sidfmd_categories ON sidfmd_categories.id=sidfmd_categories_hosts.sidfmd_category_id
         WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id AND
               sidfmd_categories.is_system IS TRUE
         ORDER BY sidfmd_categories.id
         LIMIT 1)
  GROUP BY sidfmd_issues.id
  ORDER BY sidfmd_issues.id) AS max_sris
 WHERE max_sris.max_sri >= #{opts[:corresp_required][:sri]}
 LIMIT 1
) AS not_fixed_sri")
      .select("
(
 SELECT COUNT(id)
 FROM (
  SELECT sidfmd_issues.id,MAX(COALESCE(c3_base_score,c2_base_score,0.0)) AS max_cvss_score
  FROM sidfmd_issues
  INNER JOIN issues ON issues.id=sidfmd_issues.issue_id
  LEFT OUTER JOIN sidfmd_issue_cves
   ON sidfmd_issue_cves.sidfmd_issue_id=sidfmd_issues.id
  WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
        issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
  GROUP BY sidfmd_issues.id
  ORDER BY sidfmd_issues.id
 ) AS max_cvss_scores
 WHERE max_cvss_scores.max_cvss_score >= #{opts[:corresp_required][:cvss]}
 LIMIT 1
) AS not_fixed_cvss")
      .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT (CASE WHEN issues.created_on IS NULL
              THEN ''
              ELSE TO_CHAR(issues.created_on, 'YYYY/mm/dd HH24:MI:SS OF') END)
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
),',')) AS elapsed_created_ons")
    .select("
(ARRAY_TO_STRING(ARRAY(
 SELECT
  CAST((
   SELECT issues.status_id
   FROM sidfmd_issues
   INNER JOIN issues
    ON issues.id=sidfmd_issues.issue_id
   WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
         base_sidfmd_issues.id=sidfmd_issues.id
   ORDER BY sidfmd_issues.id
   LIMIT 1
  ) AS text) || ':' ||
  CAST(
   COALESCE((
    SELECT MAX(COALESCE(c3_base_score,c2_base_score,0.0))
    FROM sidfmd_issues
    LEFT OUTER JOIN sidfmd_issue_cves
     ON sidfmd_issue_cves.sidfmd_issue_id=sidfmd_issues.id
    WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
          base_sidfmd_issues.id=sidfmd_issues.id
    GROUP BY sidfmd_issues.id
    ORDER BY sidfmd_issues.id
    LIMIT 1
   ),null) AS text) || ':' ||
   CAST(COALESCE((
    SELECT sidfmd_issue_sris.risk_id
    FROM sidfmd_issue_sris
    INNER JOIN sidfmd_issues
     ON sidfmd_issues.id=sidfmd_issue_sris.sidfmd_issue_id
    WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND
          category_id=
           (SELECT sidfmd_category_id
            FROM sidfmd_categories_hosts
            INNER JOIN sidfmd_categories
              ON sidfmd_categories.id=sidfmd_categories_hosts.sidfmd_category_id
            WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id AND
                  sidfmd_categories.is_system IS TRUE
            ORDER BY sidfmd_categories.id
            LIMIT 1) AND
          base_sidfmd_issues.id=sidfmd_issues.id
    ORDER BY sidfmd_issues.id
    LIMIT 1
   ),base_sidfmd_issues.risk_id,0) AS text)
 FROM sidfmd_issues AS base_sidfmd_issues
 WHERE base_sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id
),',')) AS sri_cvss_status_ids_orig")
    .select("
(
 SELECT COUNT(sidfmd_issues.id) > 0
 FROM sidfmd_issues
 INNER JOIN issues
  ON issues.id=sidfmd_issues.issue_id
 WHERE sidfmd_hosts.id=sidfmd_issues.sidfmd_host_id AND
       sidfmd_issues.html_id < 0 AND
       issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})
 LIMIT 1
) AS not_fixed_win")
    hosts.each do |host|
      sri_cvss_status_ids = host[:sri_cvss_status_ids_orig].split(/,/, -1).collect {|x|
        y = x.split(/:/, -1)
        { :status_id => y[0].to_i,
          :cvss => y[1] ? y[1].to_f : nil,
          :sri => y[2] ? y[2].to_i : 0,
        }
      }
      item = {
        :kb => (host[:kb] ? host[:kb] > 0 : false),
        :sri_cvss_status_ids => sri_cvss_status_ids,
        :sidfmd_host_id => host[:id],
        :hostname => host[:hostname],
        :last_checked_at => host[:last_checked_at],
        :os => host.get_os,
        :exploitable => host[:exploitable],
        :cvss => host[:cvss].split(/,/).collect {|x|
          if /\A(true|false):(\d+(\.\d+)?)\z/ =~ x then
            exploitable = (Regexp.last_match(1) == 'true')
            score = Regexp.last_match(2).to_f
            label = SidfmdIssueCve.labelling(score)
            { :expoitable => exploitable, :score => score, :label => label }
          else
            { :label => :d1, :score => 0.0, :exploitable => false }
          end
        }.compact,
        :max_sri => host[:max_sri] ? host[:max_sri] : 0,
        :softek => {
          :risk_ids => host[:risk_ids].split(/,/, -1).collect {|x| x.to_i },
          :area_ids => host[:area_ids].split(/,/, -1).collect {|x| x.to_i },
        },
        :due_dates => host[:due_dates].split(/,/, -1).collect {|x| x.size > 0 ? x : nil },
        :due_dates_unfinished => host[:due_dates_unfinished].split(/,/, -1).collect {|x| x.size > 0 ? x : nil },
        :created_ons => host[:created_ons].split(/,/, -1).collect {|x|
          x.size > 0 ? DateTime.parse(x) : DateTime.current
        },
        :tl_issue_ids => host[:tl_issue_ids].split(/,/, -1).collect {|x|
          if /\A(\d+):(\d+)?:(.*)\z/ =~ x then
            issue_id = Regexp.last_match(1).to_i
            status_id = Regexp.last_match(2).to_i
            created_on = Regexp.last_match(3)
            created_on = created_on ? DateTime.parse(created_on) : DateTime.current
            { :issue_id => issue_id, :status_id => status_id, :created_on => created_on }
          else
            nil
          end
        }.compact,
        :elapsed_created_ons => host[:elapsed_created_ons].split(/,/, -1).collect {|x|
          x.size > 0 ? DateTime.parse(x) : DateTime.current
        },
      }
      temp_sc = host[:system_category].split(/:/)
      item[:system_category] = { :id => temp_sc[0].to_i, :name => temp_sc[1] }
      item[:no_system_categories] = []
      host[:no_system_categories].split(/,/, -1).each do |cat|
        temp_cat = cat.split(/:/)
        if temp_cat.size == 2 then
          temp_id = temp_cat[0].to_i
          if opts[:role_admin] or (!opts[:role_admin] and opts[:user_category_ids].include?(temp_id)) then
            temp_name = temp_cat[1]
            item[:no_system_categories] << { :id => temp_id, :name => temp_name }
          end
        end
      end
      item[:elapsed_days] = -1
      min_elapsed_created_on = item[:elapsed_created_ons].min
      if min_elapsed_created_on then
        elapsed_days = ((Time.current - min_elapsed_created_on) / 24.0 / 60.0 / 60.0).to_i
        if item[:elapsed_days] < elapsed_days then
          item[:elapsed_days] = elapsed_days
        end
      end
      item[:max_area_id] = item[:softek][:area_ids].max
      item[:max_area_id] = 3 unless item[:max_area_id]
      item[:status_ids] = item[:sri_cvss_status_ids].collect {|x| x[:status_id] }
      item[:min_status_id] = item[:status_ids].min
      item[:min_status_id] = SidfmdIssue::STATUS_NOTAFFECT unless item[:min_status_id]
      item[:max_status_id] = item[:status_ids].max
      item[:max_status_id] = SidfmdIssue::STATUS_NOTAFFECT unless item[:max_status_id]
      item[:max_cvss] = item[:cvss].max {|x, y|
        (x[:score].nil? ? 0.0 : x[:score]) <=> (y[:score].nil? ? 0.0 : y[:score] )
      }
      if !item[:max_cvss] or item[:kb] then
        item[:max_cvss] = { :label => :notreg, :score => 0.0, :exploitable => false }
      end
      item[:warning_orig] = {
        :sri => item[:max_sri] >= opts[:corresp_required][:sri],
        :cvss => item[:max_cvss][:score] ? item[:max_cvss][:score] >= opts[:corresp_required][:cvss] : true,
      }
      item[:warning] = {
        :sri => SidfmdIssue::STATUS_UNFINISHED.include?(item[:min_status_id]) ? item[:warning_orig][:sri] : false,
        :cvss => SidfmdIssue::STATUS_UNFINISHED.include?(item[:min_status_id]) ? item[:warning_orig][:cvss] : false,
      }
      item[:fixed_warning] = {
        :sri => (SidfmdIssue::STATUS_FINISHED.include?(item[:min_status_id]) & item[:warning_orig][:sri] ),
        :cvss => (SidfmdIssue::STATUS_FINISHED.include?(item[:min_status_id]) & item[:warning_orig][:cvss]),
      }
      item[:not_fixed_warning] = {
        :sri => item[:warning][:sri],
        :cvss => item[:warning][:cvss],
      }
      ret << item
    end
    ret
  end
  
  # Agentから最後にデータを取得した日時
  def get_display_last_received_at(msg_nil: '---', format: '%Y/%m/%d %H:%M:%S')
    ret = nil
    ret = self.sidfmd_queues.order(:created_at => :desc).limit(1).first if self.is_agent
    ret ? ret.created_at.strftime(format) : msg_nil
  end

  # 脆弱性を最後に同期した日時
  def get_display_last_checked_at(msg_nil: '---', format: '%Y/%m/%d %H:%M:%S')
    ret = self.last_checked_at
    ret ? ret.strftime(format) : msg_nil
  end

  SYNCING_CHECK_SPAN = 2.0 * 24.0 * 60.0 * 60.0

  # 同期中かどうかを返す
  def is_syncing
    self.last_checked_at ? (Time.current - self.last_checked_at <= SYNCING_CHECK_SPAN) : false
  end

  # VM Agentで取得したアプリケーション情報を配列で返す。
  def get_agent_products
    queue = get_last_queue
    queue ? queue.get_agent_products : []
  end

end
