#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'delay_evaluation'

module SidfmControllerSupport

  # Exceptionをログ出力
  def log_exception(e)
    logger.error(e.message)
    if e.backtrace then
      e.backtrace.each do |l|
        logger.error(l)
      end
    end
  end

  # sidebar等、全ての画面で必要なデータを取得する。
  # 全て遅延評価で入れること
  def create_common_data
    @common = {}
    # システムカテゴリ一覧。ホストが紐づけられているもののみ。
    @common[:system_categories] = DelayEvaluation.new do
      SidfmdCategory.system.with_hosts
    end
    # システムカテゴリ数。ホストが紐づけられているもののみ。
    @common[:system_categories_size] = DelayEvaluation.new do
      @common[:system_categories].v.size
    end
    # システムカテゴリ以外のカテゴリ一覧。
    # ホストが紐づけられているもののみ。
    @common[:not_system_categories] = DelayEvaluation.new do
      if @sidfm_roles and !@sidfm_roles.include?(:admin) then
        SidfmdCategory.not_system.with_hosts.where(:id => SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id]})
      else
        SidfmdCategory.not_system.with_hosts
      end
    end
    # システムカテゴリ以外のカテゴリ数。
    # ホストが紐づけられているもののみ。
    @common[:not_system_categories_size] = DelayEvaluation.new do
      @common[:not_system_categories].v.size
    end
    # 有効な（削除されていない）ホストのOSとバージョンの情報
    @common[:dist_versions] = DelayEvaluation.new do
      SidfmdQueueDist.get_dist_versions
    end
    @common[:dist_versions_size] = DelayEvaluation.new do
      @common[:dist_versions].v.size
    end
    # 以下の状態の対策予定日付を返す。
    # ・有効な（削除されていない）ホスト
    # ・未確認、対策中のステータス
    @common[:due_dates] = DelayEvaluation.new do
      SidfmdIssue.get_due_dates
    end
    # :due_datesの数
    @common[:due_dates_size] = DelayEvaluation.new do
      @common[:due_dates].v.size
    end
    # 危険度の一覧。高い順。
    @common[:risks] = DelayEvaluation.new do
      SidfmdRisk.all.order(:id => :desc)
    end
    # 対応状況の一覧。id順（未対策->対処済)
    @common[:statuses] = DelayEvaluation.new do
      IssueStatus.all.order(:id)
    end
  end

  # CVE IDの正規化
  def normalize_cve(txt)
    ret = nil
    if /CVE-?(\d{4})-?(\d+)/i =~ txt then
      ret = "CVE-#{Regexp.last_match(1)}-#{Regexp.last_match(2)}"
    end
    ret
  end

  def display_percent(v)
    v ? "#{sprintf('%.1f', v)}%" : '---'
  end

  # 対策予定日設定済表示用
  def display_due_dates_percent(due_dates)
    return '---' unless due_dates
    f_cnt = due_dates.size
    cnt = due_dates.inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
    display_percent(cnt.to_f / f_cnt.to_f * 100.0)
  end

  # 対応状況表示用
  def display_finished_percent(status_ids)
    return '---' unless status_ids
    f_cnt = status_ids.size
    cnt = status_ids.inject(0) {|r, x|
      r += 1 if SidfmdIssue::STATUS_FINISHED.include?(x)
      r
    }
    display_percent(cnt.to_f / f_cnt.to_f * 100.0)
  end

  PROJECT_SIDFM = '脆弱性情報管理'.freeze

  # SIDfmVMのプロジェクトを取得
  def get_sidfm_project
    unless @sidfm_project then
      @sidfm_project = Project.where(:name => PROJECT_SIDFM).first
    end
    @sidfm_project
  end

  ROLE_SIDFM_ADM = 'SIDfmVM管理者'.freeze
  ROLE_SIDFM_MNG = 'SIDfmVM報告者'.freeze
  ROLE_SIDFM_USR = 'SIDfmVM閲覧者'.freeze
  ROLES_SIDFM = [ ROLE_SIDFM_ADM, ROLE_SIDFM_MNG, ROLE_SIDFM_USR ].freeze

  # SIDfmVMのカレントユーザを取得
  def get_sidfm_user
    unless @sidfm_user then
      @sidfm_user = User.current
      if @sidfm_user then
        if @sidfm_user.anonymous? then
          @sidfm_user = nil
        else
          roles = @sidfm_user.roles_for_project(get_sidfm_project)
          @sidfm_roles = []
          roles.each do |role|
            perms = role.permissions
            if perms then
              @sidfm_roles << :admin if perms.include?(:admin_sidfm)
              @sidfm_roles << :use if perms.include?(:use_sidfm)
              @sidfm_roles << :view if perms.include?(:view_sidfm)
            end
          end
          @sidfm_roles.uniq!
        end
      end
    end
    unless @sidfm_roles then
      @sidfm_roles = []
    end
    @sidfm_user
  end

  # 現在アクセス中のユーザの権限の取得
  def get_sidfm_roles
    unless @sidfm_roles then
      get_sidfm_user
    end
    @sidfm_roles
  end

  # ライセンス登録の確認
  def check_license
    if Sidfm.sidfmdlm_license_key.nil? then
      redirect_to :controller => :sidfm_licenses, :action => :license, :notice => 'ライセンス未登録です。' and return
    end
  end

  # アクセス失敗時にユーザ名をロギングする
  # 返り値は引数のmsgをそのまま返す。
  def af_log_user(msg)
    logger.info("Access Failed: #{User.current}, #{msg}")
    msg
  end

  def get_view_state
    @view_state = {}
    @view_state[:idx] = (cookies[:state_idx] || 'sri').to_sym
  end

  def put_view_state
    cookies[:state_idx] = @view_state[:idx].to_s
  end
 

  CACHE_OPTIONS = { :expires_in => 1.hour, :race_condition_ttl => 1.minute }
  
  #履歴通知機能で表示する履歴の期間
  NOTIFY_HISTORY_DATE = (Date.today - 7).to_time
 
end
