# -*- coding: utf-8 -*-
require 'date'
require 'csv'
require 'sidfm_controller_support'

class SidfmVulnerabilityController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles
  before_action :get_id

  before_action :get_view_state
  after_action :put_view_state

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_dates', 'sendmail' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
    @cache_prefix = @role_admin ? '/vul/admin/0' : "/vul/user/#{@user_category_ids.join(",")}"
  end

  # idを取得する。また、アクセスしているユーザがidに対応する脆弱性を参照可能か確認する。
  # FIXME: 複数のissueが入ることを考慮すること。
  def get_id
    @id = params[:id].to_i
    if @id == 0 and /\AKB(\d+)\z/ =~ params[:id] then
      @id = -Regexp.last_match(1).to_i
    end
    raw_issue = SidfmdIssue.enable_hosts.categorized.where(:html_id => @id)

    issue = raw_issue.count > 0 ? raw_issue.first : nil
    @html = {}
    if issue then
      @html[:html_id] = issue.html_id
      @html[:title] = issue.title
    else
      redirect_to sidfm_vulnerabilities_path, :notice => '指定された脆弱性IDは存在しません。' and return
    end
    if !@role_admin then
      flg_permit = false
      raw_issue.each do |i|
        if @user_host_ids.include?(i[:sidfmd_host_id])
          flg_permit = true
          break
        end
      end
      unless flg_permit then
        logger.error("Access Denied: user_id=#{User.current.id}, html_id=#{@id}, host_id=#{raw_issue.first[:sidfmd_host_id]}, granted_hosts=#{@user_host_ids.join(',')}")
        redirect_to sidfm_vulnerabilities_path, :notice => af_log_user('対象脆弱性に対するアクセスを許可されていません。') and return
      end
    end
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # カテゴリ選択のみ、反映させる
  def create_all_content(category_id = 0)
    @hosts[:all_content] = Rails.cache.fetch("#{@cache_prefix}/#{@id}/#{category_id}", SidfmControllerSupport::CACHE_OPTIONS) do
      get_all_content_base(category_id)
    end
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # カテゴリ選択のみ、反映させる
  def get_all_content_base(category_id)
    ret = []
    issues = SidfmdIssue
      .enable_hosts
      .joins(:issue)
      .joins(:sidfmd_host)
      .where(:html_id => @html[:html_id])
    if category_id > 0 then
      issues = issues
        .where(:sidfmd_host_id =>
               SidfmdCategoriesHost
                 .where(:sidfmd_category_id => category_id)
                 .select(:sidfmd_host_id))
    else
      issues = issues.categorized
    end
    issues = issues
      .select(:id)
      .select(:html_id)
      .select(:title)
      .select(:sidfmd_host_id)
      .select(:risk_id)
      .select(:area_id)
      .select("sidfmd_hosts.hostname AS hostname")
      .select("issues.id AS issue_id")
      .select("issues.status_id AS status_id")
      .select("TO_CHAR(issues.due_date,'YYYY/MM/DD') AS due_date")
      .select("TO_CHAR(issues.created_on, 'YYYY/MM/DD HH24:MI:SS OF') AS created_on")
      .select("(SELECT CAST(sidfmd_categories.id AS text) || ':' || sidfmd_categories.name FROM sidfmd_categories WHERE sidfmd_categories.is_system=TRUE AND sidfmd_categories.id=(SELECT sidfmd_category_id FROM sidfmd_categories_hosts WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id ORDER BY sidfmd_category_id LIMIT 1)) AS system_category")
      .select("(ARRAY_TO_STRING(ARRAY(SELECT CAST(sidfmd_categories.id AS text) || ':' || sidfmd_categories.name FROM sidfmd_categories WHERE sidfmd_categories.is_system=FALSE AND sidfmd_categories.id IN (SELECT sidfmd_category_id FROM sidfmd_categories_hosts WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id)),',')) AS no_system_categories")
      .select("(SELECT BOOL_OR(c32_exploitable) FROM (SELECT BOOL_OR(COALESCE(c3_exploitable,c2_exploitable,FALSE)) AS c32_exploitable FROM sidfmd_issue_cves WHERE sidfmd_issue_cves.sidfmd_issue_id=sidfmd_issues.id) AS c32_exploitables) AS exploitable")
      .select("(SELECT MAX(COALESCE(c3_base_score,c2_base_score)) FROM sidfmd_issue_cves WHERE sidfmd_issue_cves.sidfmd_issue_id=sidfmd_issues.id) AS cvss_score")
      .select("(SELECT sidfmd_issue_sris.risk_id FROM sidfmd_issue_sris WHERE sidfmd_issue_sris.sidfmd_issue_id=sidfmd_issues.id AND sidfmd_issue_sris.category_id=(SELECT sidfmd_category_id FROM sidfmd_categories_hosts WHERE sidfmd_categories_hosts.sidfmd_host_id=sidfmd_hosts.id ORDER BY sidfmd_category_id LIMIT 1)) AS sri")
      .select("(ARRAY_TO_STRING(ARRAY(SELECT CAST(journals.id AS text) || '!' ||  journals.journalized_type || '!' ||  CAST(journals.user_id AS text) || '!' || TO_CHAR(journals.created_on, 'YYYY/mm/dd HH24:MI:SS OF') FROM journals WHERE journals.journalized_id=issues.id AND journals.private_notes=FALSE ORDER BY journals.created_on DESC),',')) AS journals")
      .select("(SELECT TO_CHAR(MIN(issues.created_on),'YYYY/MM/DD HH24:MI:SS OF') FROM sidfmd_issues INNER JOIN issues ON issues.id=sidfmd_issues.issue_id WHERE sidfmd_issues.sidfmd_host_id=sidfmd_hosts.id AND sidfmd_issues.html_id=#{@html[:html_id]} AND issues.status_id IN (#{SidfmdIssue::STATUS_UNFINISHED.join(',')})) AS elapsed_days")
    issues.each do |issue|
      item = {
        :sidfmd_issue_id => issue[:id],
        :html_id => issue[:html_id],
        :title => issue[:title],
        :sidfmd_host_id => issue[:sidfmd_host_id],
        :hostname => issue[:hostname],
        :os => issue.sidfmd_host.get_os,
        :softek => { :risk_id => issue[:risk_id], :area_id => issue[:area_id] },
        :issue_id => issue[:issue_id],
        :status_id => issue[:status_id],
        :due_date => issue[:due_date],
        :created_on => { :value => DateTime.parse(issue[:created_on]) },
        :system_category => if /\A(\d+):(.*)\z/ =~ issue[:system_category] then
                              cat_id = Regexp.last_match(1).to_i
                              cat_name = Regexp.last_match(2)
                              { :id => cat_id, :name => cat_name }
                            else
                              nil
                            end,
        :no_system_categories => issue[:no_system_categories].split(/,/, -1).collect {|x|
          if /\A(\d+):(.*)\z/ =~ x then
            cat_id = Regexp.last_match(1).to_i
            cat_name = Regexp.last_match(2)
            if @role_admin or (!@role_admin and @user_category_ids.include?(cat_id)) then
              { :id => cat_id, :name => cat_name }
            else
              nil
            end
          else
            nil
          end
        }.compact,
        :exploitable => issue[:exploitable],
        :cvss => { :score => issue[:cvss_score] },
        :sri => issue[:sri] ? issue[:sri] : 0,
        :journals => issue[:journals].split(/,/).collect {|x|
          if /\A(\d+)!([^:]*)!(.*)!(.*)\z/ =~ x then
            journal_id = Regexp.last_match(1).to_i
            journal_type = Regexp.last_match(2)
            user_id = Regexp.last_match(3).to_i
            created_on = Regexp.last_match(4)
            created_on = DateTime.parse(created_on) if created_on
            { :id => journal_id,
              :type => journal_type,
              :user_id => user_id,
              :created_on => created_on, }
          else
            nil
          end
        }.compact,
      }
      item[:elapsed_days] = '---'
      if issue[:elapsed_days] and issue[:elapsed_days].size > 1 then
        item[:elapsed_days] = ((Time.current - DateTime.parse(issue[:elapsed_days]))/24.0/60.0/60.0).to_i
      end
      item[:created_on][:display] = item[:created_on][:value] ? item[:created_on][:value].localtime.strftime('%Y/%m/%d') : '---'
      item[:created_on][:elapsed_days] = item[:created_on][:value] ? ((Time.current - item[:created_on][:value]) / 24.0 / 60.0 / 60.0).to_i : -1
      if item[:cvss][:score] then
        item[:cvss][:label] = SidfmdIssueCve.labelling(item[:cvss][:score])
      else
        item[:cvss][:label] = :notreg
      end
      if @id < 0 then
        item[:sri] = issue[:risk_id]
      end
      item[:warning_orig] = {
        :sri => item[:sri] >= @corresp_required[:sri],
        :cvss => item[:cvss][:score] ? (item[:cvss][:score] >= @corresp_required[:cvss]) : true,
      }
      item[:warning] = {
        :sri => SidfmdIssue::STATUS_UNFINISHED.include?(item[:status_id]) ? item[:warning_orig][:sri] : false,
        :cvss => SidfmdIssue::STATUS_UNFINISHED.include?(item[:status_id]) ? item[:warning_orig][:cvss] : false,
      }
      item[:fixed_warning] = {
        :sri => (SidfmdIssue::STATUS_FINISHED.include?(item[:status_id]) & item[:warning_orig][:sri]),
        :cvss => (SidfmdIssue::STATUS_FINISHED.include?(item[:status_id]) & item[:warning_orig][:cvss]),
      }
      ret << item
    end
    ret
  end

  FILTER_KEYS = {}

  def init_filter_keys
    @dist_versions = @hosts[:all_content].collect {|x|
      (x[:os] and  x[:os][:osname] and x[:os][:osname].size > 0) ?
      x[:os][:osname] : nil
    }.compact.uniq.sort

    @due_dates = @hosts[:all_content].inject([]) {|r, x|
      r += [ x[:due_date] ]
    }.uniq.collect {|x| x.nil? ? '' : x }.sort

    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:warning] = { :values => [ 'warning', 'others', 'all' ].map(&:freeze).freeze, :default => 'all'.freeze }
      FILTER_KEYS[:status] = {
        :values => (SidfmdIssue::STATUS_ALL.collect {|x| x.to_s } +
                    [ SidfmdIssue::STATUS_UNFINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ SidfmdIssue::STATUS_FINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ '0' ]).map(&:freeze).freeze, :default => '0'.freeze }
      FILTER_KEYS[:sri] = { :values => [ '1', '2', '3', '4', '0', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:cvss] = { :values => [ 'd1', 'd2', 'd3', 'd4', 'notreg', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:risk] = { :values => [ '1', '2', '3', '4', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:area] = { :values => [ '3', '2', '1', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:exploitable] = { :values => [ 'true', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:journal] = { :values => [ 'exist', '1day', '1week', '1month', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:os] = { :values => @dist_versions + [ ''.freeze ], :default => '' }
      FILTER_KEYS[:due_date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:keyword] = { :values => nil, :default => '' }
    end
  end

  def check_filters
    init_filter_keys
    f_keys = FILTER_KEYS.keys
    for i in 0..f_keys.size - 1 do
      if @hosts[:filters].has_key?(f_keys[i]) then
        if FILTER_KEYS[f_keys[i]][:values] && !FILTER_KEYS[f_keys[i]][:values].include?(@hosts[:filters][f_keys[i]]) then
          @hosts[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
        end
      else
        @hosts[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
      end
    end
  end

  DAY_SECONDS = 24 * 60 * 60
  DATE_INVALID = Date.new(4000,12,31).freeze
  DATE_NONE = Date.new(3000,12,31).freeze

  # フィルタの適用
  def apply_filters
    ret = @hosts[:all_content].find_all {|item|
      # 要注意
      if @hosts[:filters][:warning] == 'warning' then
        next false unless item[:warning][@hosts[:idx]]
      elsif @hosts[:filters][:warning] == 'others' then
        next false if item[:warning][@hosts[:idx]]
      end
      # 対応状況
      if /\A(\d):(\d)\z/ =~ @hosts[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:status_id] == status1 or item[:status_id] == status2)
      elsif @hosts[:filters][:status] != '0' then
        next false unless item[:status_id] == @hosts[:filters][:status].to_i
      end
      # SRI
      if @hosts[:filters][:sri] != '' then
        next false unless item[:sri] == @hosts[:filters][:sri].to_i
      end
      # CVSS
      if @hosts[:filters][:cvss] != '' then
        next false unless item[:cvss][:label] == @hosts[:filters][:cvss].to_sym
      end
      # 影響範囲
      if @hosts[:filters][:area] != '' then
        next false unless (item[:html_id] > 0 and @hosts[:filters][:area].to_i == item[:softek][:area_id])
      end
      # 攻撃コード
      if @hosts[:filters][:exploitable] != '' then
        next false unless (@hosts[:filters][:exploitable] == 'true' && item[:exploitable])
      end
      # OS
      if @hosts[:filters][:os] != '' then
        next false unless (item[:os] and @hosts[:filters][:os] == item[:os][:osname])
      end
      # 履歴
      if @hosts[:filters][:journal] != '' then
        if @hosts[:filters][:journal] == 'exist' then
          next false unless item[:journals].size > 0 
        else
          temp_timespan = 1.days
          case @hosts[:filters][:journal]
          when '1day'   then temp_timespan = 1.days
          when '1week'  then temp_timespan = 1.weeks
          when '1month' then temp_timespan = 1.months
          end
          timespan = Time.current.ago(temp_timespan)
          next false unless item[:journals].find {|x|
            x[:created_on] >= timespan
          }
        end
      end
      # 対策予定日設定済
      if @hosts[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @hosts[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless (item[:due_date].nil? or item[:due_date].size == 0)
          else
            next false if item[:due_date].nil?
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              log_exception(e)
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              log_exception(e)
              enddate = nil
            end
            due_date = nil
            begin
              due_date = Date.parse(item[:due_date])
            rescue Exception => e
              log_exception(e)
              due_date = nil
            end
            if begindate and enddate then
              next false unless (due_date >= begindate and due_date <= enddate)
            elsif begindate then
              next false unless due_date >= begindate
            elsif enddate then
              next false unless due_date <= enddate
            end
          end
        end
      end
      # キーワード検索
      if @hosts[:filters][:keyword].size > 0 then
        next false unless
          Regexp.new('.*' + Regexp.escape(@hosts[:filters][:keyword]) + '.*',
                     Regexp::IGNORECASE) =~ item[:title] or
          Regexp.new('.*' + Regexp.escape(@hosts[:filters][:keyword]) + '.*',
                     Regexp::IGNORECASE) =~ item[:hostname]
      end
      # 日付指定検索
      if @hosts[:filters][:date] != '::' then
        if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @hosts[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @hosts[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false if begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false if end_date >= item[:created_on][:value]
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless begin_date <= item[:created_on][:value] and end_date >= item[:created_on][:value]
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless begin_date <= item[:created_on][:value]
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless end_date >= item[:created_on][:value]
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          end
        else
          @hosts[:filters][:date] = '::'
          next false
        end
      end
      true
    }
    #ret.sort! {|x, y|
    #  x[:status_id] != y[:status_id] ?
    #  SidfmdIssue::SORT_PRIORITIES[x[:status_id]] <=> SidfmdIssue::SORT_PRIORITIES[y[:status_id]] :
    #  y[:created_on][:value] <=> x[:created_on][:value]
    #}
    case @hosts[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:warning][@hosts[:idx]] ? 500 : 100 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:warning][@hosts[:idx]] ? 500 : 100 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'status_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:status_id] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'status_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:status_id] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'sri_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:sri] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'sri_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:sri] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'cvss_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:cvss][:score] ? x[:cvss][:score] : 0.0 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'cvss_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:cvss][:score] ? x[:cvss][:score] : 0.0 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'area_asc' then
      temp_ev = proc {|x|
        [
         x[:hostname],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'area_desc' then
      temp_ev = proc {|x|
        [
         x[:hostname],
         x[:html_id] > 0 ? x[:softek][:area_id] : -100,
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [
         x[:hostname],
         x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 25) : 40
        ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [
         x[:hostname],
         x[:html_id] > 0 ? (x[:cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 25) : 40
        ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'elapseddays_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:elapsed_days] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'elapseddays_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:elapsed_days] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'hostname_asc' then
      ret.sort! {|x, y| x[:hostname].downcase <=> y[:hostname].downcase }
    when 'hostname_desc' then
      ret.sort! {|y, x| x[:hostname].downcase <=> y[:hostname].downcase }
    when 'categories_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'categories_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'os_asc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:hostname], osname, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'os_desc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:hostname], osname, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'journals_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:journals].size ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'journals_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:journals].size ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'duedate_asc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:due_date] ? x[:due_date] : '0000/00/00' ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'duedate_desc' then
      temp_ev = proc {|x|
        [ x[:hostname], x[:due_date] ? x[:due_date] : '0000/00/00' ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    end
    ret
  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month

  # タイムラインのデータを返す
  def get_timeline_data
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @hosts[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    @hosts[:all_content].each do |item|
      ret[:timescale].each do |step|
        next if item[:created_on][:value] > step[:time]
        # issue
        issue_status_id = item[:status_id]
        # journals
        temp_j = item[:journals].collect {|x|
          if x[:created_on] <= step[:time] and x[:details] and
              x[:details].find {|x| x[:prop_key] == 'status_id' } then
            x
          else
            nil
          end
        }
        temp_j.compact!
        temp_j.sort! {|x, y|
          x[:created_on] == y[:created_on] ?
          x[:id] <=> y[:id] :
          x[:created_on] <=> y[:created_on] }
        journal_status_id = nil
        if temp_j.size > 0 then
          temp_j.last[:details].each do |i|
            if i[:prop_key] == 'status_id' then
              journal_status_id = i[:value].to_i
              break
            end
          end
        end
        count_index = TIMELINE_SPAN - 1 - step[:step]
        if journal_status_id then
          ret[:status_counts][journal_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(journal_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        elsif issue_status_id then
          ret[:status_counts][issue_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(issue_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          else
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][5][count_index] += 1
            when :cvss then
              ret[:risk_counts][:fixed][count_index] += 1
            end
          end
        end
      end
    end

    ret
  end

  def get_timeline_data_orig
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @hosts[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1 ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    @hosts[:all_content].each do |item|
      ret[:timescale].each do |step|
        next if item[:created_on][:value] > step[:time]
        # issue
        issue_status_id = item[:status_id]
        # journals
        temp_j = item[:journals].collect {|x|
          if x[:created_on] <= step[:time] and x[:details] and
              x[:details].find {|x| x[:prop_key] == 'status_id' } then
            x
          else
            nil
          end
        }
        temp_j.compact!
        temp_j.sort! {|x, y|
          x[:created_on] == y[:created_on] ?
          x[:id] <=> y[:id] :
          x[:created_on] <=> y[:created_on] }
        journal_status_id = nil
        if temp_j.size > 0 then
          temp_j.last[:details].each do |i|
            if i[:prop_key] == 'status_id' then
              journal_status_id = i[:value].to_i
              break
            end
          end
        end
        count_index = TIMELINE_SPAN - 1 - step[:step]
        if journal_status_id then
          ret[:status_counts][journal_status_id][count_index] += 1
          if SidfmdIssue::STATUS_UNFINISHED.include?(journal_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          end
        elsif issue_status_id then
          ret[:status_counts][issue_status_id][count_index] += 1
          #if SidfmdIssue::STATUS_UNFINISHED.include?(issue_status_id) then
            case @hosts[:idx]
            when :sri then
              ret[:risk_counts][item[:sri]][count_index] += 1
            when :cvss then
              ret[:risk_counts][item[:cvss][:label]][count_index] += 1
            end
          #end
        end
      end
    end

    ret
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :grp, :idx, :category_id, :filters, :checked, :all_checked)

    @hosts = { :url => request.url }
    @system_categories = SidfmdCategory.system
    @not_system_categories = SidfmdCategory.not_system
    if !@role_admin then
      @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @hosts[:category_id] = (pages_params[:category_id] || 0).to_i
    @hosts[:category_name] = '全てのホスト'
    if @hosts[:category_id] > 0 and
        (@system_categories.where(:id => @hosts[:category_id]).size > 0 or
         @role_admin or
         (!@role_admin and @user_category_ids.include?(@hosts[:category_id]))) then
      cat = SidfmdCategory.find(@hosts[:category_id])
      @hosts[:category_name] = cat.name if cat
    else
      @hosts[:category_id] = 0
    end

    @hosts[:grp] = (pages_params[:grp] || 'count').to_sym
    @hosts[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @hosts[:idx]
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 10).to_i
    @hosts[:order] = pages_params[:order] || 'hostname_asc'
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = (pages_params[:all_checked] || '').to_s
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys

    @corresp_required = Sidfm.corresp_required
    create_all_content(@hosts[:category_id])
    check_filters


    @hosts[:corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:warning_orig][@hosts[:idx]] ? 1 : 0 }
    @hosts[:fixed_corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:fixed_warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:not_fixed_corresp_required_host_size] =
      @hosts[:corresp_required_host_size] - @hosts[:fixed_corresp_required_host_size]
    @hosts[:all_host_size] = @hosts[:all_content].size
  end

  def filter_content
    @status = {
      :exploit => [ { :value => @hosts[:not_fixed_corresp_required_host_size] } ]
    }
    @hosts[:distribution_status] = {
      :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0,
    }
    @hosts[:distribution_risk] =
      case @hosts[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @hosts[:all_content].each do |i|
      case i[:status_id]
      when SidfmdIssue::STATUS_NOTCONFIRMED then
        @hosts[:distribution_status][:notconfirmed] += 1
      when SidfmdIssue::STATUS_ONGOING then
        @hosts[:distribution_status][:ongoing] += 1
      when SidfmdIssue::STATUS_FIXED then
        @hosts[:distribution_status][:fixed] += 1
      when SidfmdIssue::STATUS_NOTAFFECT then
        @hosts[:distribution_status][:notaffect] += 1
      end
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:status_id]) then
        case @hosts[:idx]
        when :sri then
          case i[:sri]
          when SidfmdIssueSri::VALUES[0] then
            @hosts[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @hosts[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @hosts[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @hosts[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @hosts[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:cvss][:label]
          when :notreg then
            @hosts[:distribution_risk][:notreg] += 1
          when :d4 then
            @hosts[:distribution_risk][:d4] += 1
          when :d3 then
            @hosts[:distribution_risk][:d3] += 1
          when :d2 then
            @hosts[:distribution_risk][:d2] += 1
          when :d1 then
            @hosts[:distribution_risk][:d1] += 1
          end
        end
      else
        @hosts[:distribution_risk][:fixed] += 1
      end
    end
    @hosts[:content_filtered] = apply_filters
    @hosts[:total_size] = @hosts[:content_filtered].size

    @hosts[:statuses] = SidfmdIssue.statuses

    @statuses = IssueStatus.all.order(:id)

    @hosts[:dist_cat_hosts] = {}
    [ :public_server, :private_server, :shared_server,
      :mobile, :client, :virtual_host_server, :public_network,
      :private_network ].each_with_index do |v, i|
      @hosts[:dist_cat_hosts][v] =
        @hosts[:all_content].sum {|x| x[:system_category][:id] == i + 1 ? 1 : 0 }
    end
    @hosts[:dist_os_hosts] = {}
    @hosts[:all_content].each do |host|
      os =
        (host[:os] and  host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '（不明）'
      if @hosts[:dist_os_hosts].has_key?(os) then
        @hosts[:dist_os_hosts][os] += 1
      else
        @hosts[:dist_os_hosts][os] = 1
      end
    end

    @timeline_data = (@hosts[:grp] == :distribution) ? get_timeline_data : []
  end

  public

  def index
    f_params = params.fetch(:form, {}).permit(:history, :is_update_one_row)
    parse_params
    filter_content
    if session['hostid_with_htmlid'] then
      if session['hostid_with_htmlid']['host'] then
        @hosts[:content_filtered].each do |host|
          h_id = host[:sidfmd_host_id].to_s
          if session['hostid_with_htmlid']['host'][h_id] then
            session['hostid_with_htmlid']['host'][h_id][1] = true
          end
        end
      end
      if session['hostid_with_htmlid']['vul'] && !f_params[:is_update_one_row] then
        session['hostid_with_htmlid']['vul'].each_key do |htmlid|
          session['hostid_with_htmlid']['vul'][htmlid][1] = false
        end
      end
    end
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    if f_params[:history] then
      history = JSON.parse(f_params[:history])
      host_id = history[0].to_i
      html_id = history[1].to_i
      count = 0
      @hosts[:content_filtered].each do |host|
        if host[:sidfmd_host_id] == host_id then
          break
        else
          count += 1;
        end
      end
      @hosts[:offset] = count
      @history = host_id
      session['delete_histories'] = [host_id, html_id]
    end
    @hosts[:content] = @hosts[:content_filtered].slice(@hosts[:offset], @hosts[:per_page])
    @hosts[:content] = [] unless @hosts[:content]
   
    #ブラウザバックかどうか判断するための処理
    if !session['chkbok_ids'] then
      session['chkbok_ids'] = {}
    end
    if !session['chkbok_ids']['vul'] then
      session['chkbok_ids']['vul'] = {}
    end
    session['chkbok_ids']['vul']['is_browser_back'] = false
  end

  def update_status
    f_params = params.fetch(:status, []).permit(:hostids, :prev_status, :new_status, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !hostids.respond_to?(:'[]') or !hostids.respond_to?(:size) or hostids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定後ステータスが不正です。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if new_status == 0 then
      dump[:notice] = '設定後ステータスが設定されなかったため、操作は行われておりません。'
      redirect_to dump and return
    end
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
          .where(:html_id => @id)
        if prev_status == 0 then
          issues = issues
            .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        else
          issues = issues
            .where(:issues => { :status_id => prev_status })
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_status_with_journal(User.current.id, new_status, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end
    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end

    redirect_to dump
  end

  def update_due_dates
    f_params = params.fetch(:schedule, {}).permit(:hostids, :force, :due_date, :comment, :dump, :action)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if hostids.size == 0 or !hostids.respond_to?(:'[]') then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
          .where(:html_id => @id)
        if !force then
          issues = issues
            .where('issues.due_date IS NULL')
        end
        issues.each do |issue|
          issue.with_lock do
            issue.change_due_date_with_journal(User.current.id, due_date, comment)
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    action = f_params[:action]
    if action == 'single' then
      if !session['hostid_with_htmlid'] then
        session['hostid_with_htmlid'] = {}
      end
      if !session['hostid_with_htmlid']['vul'] then
        session['hostid_with_htmlid']['vul'] = {}
      end
      session['hostid_with_htmlid']['vul'][@id.to_s] = [hostids[0].to_s, true]
      dump[:'form[is_update_one_row]'] = true
    end

    redirect_to dump
  end

  def sendmail
    f_params = params.fetch(:sm, {}).permit(:title, :body, :htmlids)
    logger.info(f_params)
    redirect_to :action => :index
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  def download
    pages_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :hostids)
    format = pages_params[:format]
    if ![ 'json', 'tsv' ].include?(format) then
      redirect_to :action => :index, :notice => '不正な入力です。' and return
    end
    @hosts = { :url => request.url }
    @hosts[:category_id] = (pages_params[:category_id] || 0).to_i
    @hosts[:category_name] = '全てのホスト'
    if @hosts[:category_id] > 0 then
      cat = SidfmdCategory.find(@hosts[:category_id])
      @hosts[:category_name] = cat.name if cat
    end
    @hosts[:idx] = (pages_params[:idx] || 'sri').to_sym
    @hosts[:order] = (pages_params[:order] || 'id').to_sym
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    @hosts[:hostids] = []
    begin
      @hosts[:hostids] = JSON.parse(pages_params[:hostids] || '[]')
    rescue JSON::ParserError => e
      logger.error(e)
    end
    @corresp_required = Sidfm.corresp_required
    create_all_content(@hosts[:category_id])
    check_filters
    @hosts[:corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:fixed_corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:fixed_warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:not_fixed_corresp_required_host_size] =
      @hosts[:corresp_required_host_size] - @hosts[:fixed_corresp_required_host_size]
    @hosts[:all_host_size] = @hosts[:all_content].size
    filter_content
    case format
    when 'json' then
      json = {}
      temp_order = @hosts[:order].to_s.split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @hosts[:filters]
      json[:category] = { :id => @hosts[:category_id], :name => @hosts[:category_name], }
      json[:idx] = @hosts[:idx]
      json[:content] = []
      @hosts[:content_filtered].each do |host|
        if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
          next
        end
        item = {}
        item[:host] = { :id => host[:sidfmd_host_id], :hostname => host[:hostname], }
        item[:warning] = host[:warning][@hosts[:idx]]
        item[:status] = SidfmdIssue.status_id_to_s(host[:status_id])
        item[:sri] = SidfmdIssueSri::RISKS[host[:sri]]
        t_cvss = host[:cvss]
        item[:cvss] = (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
        item[:area] =  SidfmdArea.where(:area_id => host[:softek][:area_id]).first.name
        item[:exploitable] = (t_cvss[:label] != :notreg ? host[:exploitable] : nil)
        cats = []
        cats << host[:system_category][:name]
        host[:no_system_categories].each do |cat|
          cats << cat[:name]
        end
        item[:categories] = cats
        item[:os] = ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '')
        item[:journals] =  []
        host[:journals].each do |journal|
          j_item = {}
          j_item[:id] = journal[:id]
          j_item[:created_on]  = journal[:created_on] ? journal[:created_on].strftime('%Y/%m/%d %H:%M:%S') : '---'
          j_item[:user] = { :id => journal[:user_id], :name => view_context.display_user_name(User.find(journal[:user_id])) }
          j_item[:notes] = Journal.find(journal[:id]).notes
          j_item[:details] = []
          JournalDetail.where(:journal_id => journal[:id]).order(:id).each do |detail|
            jd_item = {}
            jd_item[:id] = detail[:id]
            case detail[:prop_key]
            when 'status_id' then
              jd_item[:action] = 'ステータス変更'
              jd_item[:from] = SidfmdIssue.status_id_to_s(detail[:old_value].to_i)
              jd_item[:to] = SidfmdIssue.status_id_to_s(detail[:value].to_i)
            when 'due_date' then
              jd_item[:action] = '対策予定日の変更'
              jd_item[:from] = SidfmdIssue.due_date_to_s(detail[:old_value])
              jd_item[:to] = SidfmdIssue.due_date_to_s(detail[:value])
            end
            j_item[:details] << jd_item
          end
          item[:journals] << j_item
        end
        item[:due_date] = host[:due_date]
        apps = []
        pkgs = []
        happs = []
        sidfmd_host = SidfmdHost.find(host[:sidfmd_host_id])
        if sidfmd_host then
          queue = sidfmd_host.get_last_queue
          if queue then
            pkgs = queue.sidfmd_queue_pkgs.order(:pkg).collect {|x| x.pkg }
            apps = queue.sidfmd_queue_apps.order(:name,:version).collect {|x| {
                :name => x.name, :version => x.version, :path => x.path,
              }}
          end
          happs = sidfmd_host.sidfmd_host_apps.order(:product_id,:version_id).collect {|x| x.get_name }
        end
        item[:packages] = pkgs
        item[:products] = apps
        item[:manual_products] = happs

        item[:affect_packages] = []
        item[:affect_products] = []

        item[:issue] = { :id => host[:sidfmd_issue_id] }
        issue = SidfmdIssue.find(host[:sidfmd_issue_id])
        if issue then
          item[:affect_packages] = issue.affect_packages('')
          item[:affect_products] = issue.affect_products
          item[:issue][:title] = issue.title
          item[:issue][:html_id] = issue.html_id > 0 ? "#{issue.html_id}" : "KB#{-issue.html_id}"
        end
        json[:content] << item
      end
      send_data(json.to_json, :filename => "vmlists-vulnerability#{@id}-json-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
      return
    when 'tsv' then
      header = [ 'ホストID', '要対応', '対応状況', 'SRI', 'CVSS', '影響範囲', '攻撃コード', 'ホスト名', 'カテゴリ', 'OS名', '履歴', '対策予定日' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          item.clear
          item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
          item << host[:warning][@hosts[:idx]].to_s.encode('UTF-16LE')
          item << SidfmdIssue.status_id_to_s(host[:status_id]).encode('UTF-16LE')
          item << SidfmdIssueSri::RISKS[host[:sri]].encode('UTF-16LE')
          t_cvss = host[:cvss]
          item << (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => host[:softek][:area_id]).first.name.encode('UTF-16LE')
          item << host[:exploitable].to_s.encode('UTF-16LE')
          item << host[:hostname].encode('UTF-16LE')
          cats = []
          cats << host[:system_category][:name].encode('UTF-16LE')
          host[:no_system_categories].each do |cat|
            cats << cat[:name].encode('UTF-16LE')
          end
          item << cats.join(','.encode('UTF-16LE'))
          item << ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '').encode('UTF-16LE')
          item << host[:journals].size.to_s.encode('UTF-16LE')
          item << (host[:due_date] ? host[:due_date] : '').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlists-vulnerability#{@id}-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end

    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end
  
  def send_mail
    filtered_params = params.fetch(:form, {}).permit(:recipients, :subject, :message)
  begin
    mail = {}
    mail[:from] = User.current.mail
    email_with_name = []
    recipients = JSON.parse(filtered_params[:recipients])
    recipients.each do |key, val|
      email_name = %("#{val}" <#{key}>)
      email_with_name << email_name
    end
    mail[:to] = email_with_name
    mail[:subject] = filtered_params[:subject]
    mail[:message] = filtered_params[:message]
    SidfmdMailer.send_mail(mail).deliver_now
  rescue Exception => e
    log_exception(e)
    redirect_to :action => :index, :notice => e and return
  end
    redirect_to :action => :index and return
  end

end

